# SPDX-FileCopyrightText:  2024 - Wei Zhao <wei.zhao@uclouvain.be>
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.model import ModelView
from trytond.wizard import Wizard, StateView, StateTransition, Button
from trytond.pool import Pool
import logging

__all__ = ['GetNewStudiesStart', 'GetNewStudies']

logger = logging.getLogger(__name__)


class GetNewStudiesStart(ModelView):
    """
    Get New Studies Start
    """
    __name__ = "gnuhealth.imaging_orthanc.get_new_studies.start"

#
# This class is responsible for retrieving
# all saved studies from the Orthanc server.


class GetNewStudies(Wizard):
    "Get New Studies"
    __name__ = 'gnuhealth.imaging_orthanc.get_new_studies'

    start = StateView(
        'gnuhealth.imaging_orthanc.get_new_studies.start',
        'health_orthanc.get_new_studies_start_form',
        [Button("Cancel", 'end', 'tryton-cancel'),
         Button("Start", 'update', 'tryton-ok'),
         ])
    update = StateTransition()

    def transition_update(self):
        """
        Get new studies and return 'end'.
        """
        Pool().get('gnuhealth.imaging_orthanc.study').get_new_studies()
        return 'end'

    def end(self):
        """
        Method to signal the end of the process and return the string 'reload'.
        """
        return 'reload'



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                        HEALTH LAB package                             #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import health_lab
from . import wizard
from . import sequences


def register():
    Pool.register(
        health_lab.PatientData,
        health_lab.TestType,
        health_lab.Lab,
        health_lab.GnuHealthLabTestUnits,
        health_lab.GnuHealthTestCritearea,
        health_lab.GnuHealthPatientLabTest,
        wizard.wizard_create_lab_test.CreateLabTestOrderInit,
        wizard.wizard_create_lab_test.RequestPatientLabTestStart,
        wizard.wizard_create_lab_test.RequestTest,
        health_lab.PatientHealthCondition,
        sequences.GnuHealthSequences,
        sequences.LabRequestSequence,
        sequences.LabTestSequence,
        module='health_lab', type_='model')

    Pool.register(
        wizard.wizard_create_lab_test.CreateLabTestOrder,
        wizard.wizard_create_lab_test.RequestPatientLabTest,
        module='health_lab', type_='wizard')



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH LAB package                                #
#                health_lab.py: main module                             #
#########################################################################
from datetime import datetime
from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.pool import Pool, PoolMeta
from trytond.pyson import Eval, Not, Bool, Equal, If
from trytond.modules.health.core import (get_health_professional,
                                         get_age_for_comparison)

import re
from uuid import uuid4

try:
    from PIL import Image
except ImportError:
    Image = None


__all__ = [
    'PatientData', 'TestType', 'Lab',
    'GnuHealthLabTestUnits', 'GnuHealthTestCritearea',
    'GnuHealthPatientLabTest', 'PatientHealthCondition']


class PatientData(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient'

    lab_test_ids = fields.One2Many(
        'gnuhealth.patient.lab.test', 'patient_id',
        'Lab Tests Required')


class TestType(ModelSQL, ModelView):
    'Type of Lab test'
    __name__ = 'gnuhealth.lab.test_type'

    name = fields.Char(
        'Test',
        help="Test type, eg X-Ray, hemogram,biopsy...", required=True,
        translate=True)
    code = fields.Char(
        'Code',
        help="Short name - code for the test", required=True)

    specimen_type = fields.Char(
        'Specimen',
        help='Specimen type, for example: '
        'whole blood, plasma, urine or feces ...',
        translate=True)

    info = fields.Text('Description')
    product_id = fields.Many2One('product.product', 'Service', required=True)
    critearea = fields.One2Many(
        'gnuhealth.lab.test.critearea', 'test_type_id',
        'Test Cases')

    gender = fields.Selection([
        (None, ''),
        ('m', 'Male'),
        ('f', 'Female'),
    ], 'Gender')

    @staticmethod
    def default_gender():
        return None

    min_age = fields.Float(
        "Min age",
        help='Min age year, '
        '(years x 365 + months x 30.5 + days) / 365')

    @staticmethod
    def default_min_age():
        return 0

    max_age = fields.Float(
        "Max age",
        help='Max age year, '
        '(years x 365 + months x 30.5 + days) / 365')

    @staticmethod
    def default_max_age():
        return 150

    age_range = fields.Function(
        fields.Char('Age range'), 'get_age_range')

    def get_age_range(self, name):
        min_age = self.min_age
        max_age = self.max_age
        if min_age is not None and max_age is not None:
            if min_age is None:
                min_age = 0
            if max_age is None:
                max_age = 150
            return str(min_age) + "-" + str(max_age)

    category = fields.Selection([
        (None, ''),
        ('hematology', 'Hematology Testing'),
        ('fluid_excreta', 'Body Fluid and Excreta Examination'),
        ('biochemical', 'Biochemical Testing'),
        ('immunological', 'Immunological Testing'),
        ('microbiological', 'Microbiological Testing'),
        ('molecular_biology', 'Molecular Biology Testing'),
        ('chromosome_genetic', 'Chromosome and Genetic Disease Detection'),
        ('others', 'Others'),
    ], 'Category', sort=False)

    @staticmethod
    def default_category():
        return None

    report_style = fields.Selection([
        ('tbl_h_r_u_nr',
         'Table with result, unit and normal_range columns'),
        ('tbl_h_r_nr', 'Table with result and normal_range columns'),
        ('tbl_h_r', 'Table with result column'),
        ('tbl_nh_r', 'Table with result column (no header)'),
        ('tbl_nh_r_img',
         'Table with result column and inline images (no header)'),
        ('no_tbl', 'Do not use table'),
        ('do_not_show', 'Do not show in report'),
    ], 'Report style', sort=False)

    @staticmethod
    def default_report_style():
        return 'tbl_h_r_nr'

    tags = fields.Char(
        'Tags', help='Tags of test type, which can be used in '
        'if directive of report template file, '
        'tags use letters and numbers, separated by colon.')

    # Mostly used in report template file.
    def all_tags(self):
        tags = self.tags.split(':')
        return tags.sort()

    # Mostly used in report template file.
    def has_tag(self, tag):
        tags = self.tags.split(':')
        return (tag in tags)

    active = fields.Boolean('Active')

    @staticmethod
    def default_active():
        return True

    @classmethod
    def __setup__(cls):
        super(TestType, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('code_uniq', Unique(t, t.name),
             'The Lab Test code must be unique')
        ]

        cls._order.insert(0, ('category', 'ASC'))
        cls._order.insert(1, ('name', 'ASC'))
        cls._order.insert(2, ('gender', 'ASC'))
        cls._order.insert(3, ('min_age', 'ASC'))
        cls._order.insert(4, ('max_age', 'ASC'))
        cls._order.insert(4, ('tags', 'ASC'))

    @classmethod
    def check_xml_record(cls, records, values):
        return True

    @classmethod
    def search_rec_name(cls, name, clause):
        """ Search for the full name and the code """
        field = None
        for field in ('name', 'code'):
            tests = cls.search([(field,) + tuple(clause[1:])], limit=1)
            if tests:
                break
        if tests:
            return [(field,) + tuple(clause[1:])]
        return [(cls._rec_name,) + tuple(clause[1:])]

    @classmethod
    def write(cls, test_types, values):
        for test_type in test_types:
            if values.get('tags') != '':
                tags = values.get('tags', '').split(':')
                tags = [re.sub(r'[^\w_@#%]', '', tag) for tag in tags]
                tags = sorted(set([tag for tag in tags if tag != '']))
                values['tags'] = ":".join(tags)
        return super(TestType, cls).write(test_types, values)


class Lab(ModelSQL, ModelView):
    'Patient Lab Test Results'
    __name__ = 'gnuhealth.lab'

    STATES = {'readonly': Eval('state') == 'validated'}

    name = fields.Char('ID', help="Lab result ID", readonly=True)
    test = fields.Many2One(
        'gnuhealth.lab.test_type', 'Test type',
        help="Lab test type", required=True)
    source_type = fields.Selection([
        ('patient', 'Patient'),
        ('other_source', 'Other')
    ], 'Source',
        help='Sample source type.',
        sort=False)
    source_type_str = source_type.translated('source_type')

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient',
        states={'invisible': (Eval('source_type') != 'patient')},
        help="Patient")

    other_source = fields.Char(
        'Other',
        states={
            'invisible': (
                Eval('source_type') != 'other_source')},
        help="Other sample source.")
    source_name = fields.Function(
        fields.Text('Source name'), 'get_source_name')

    def get_source_name(self, name=None, with_puid=False, with_gender=False):
        if self.is_patient():
            pname = (self.patient
                     and self.patient.rec_name
                     or '')
            puid_str = (with_puid
                        and self.patient
                        and f' ({self.patient.puid})'
                        or '')
            gender_str = (with_gender
                          and self.patient
                          and f' [{self.patient.gender_str}]'
                          or '')
            return pname + puid_str + gender_str
        else:
            return (self.other_source or '')

    specimen_type = fields.Char(
        'Specimen',
        help='Specimen type, for example: '
        'whole blood, plasma, urine or feces ...')

    pathologist = fields.Many2One(
        'gnuhealth.healthprofessional', 'Pathologist',
        help="Pathologist")
    requestor = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof',
        help="Doctor who requested the test")
    results = fields.Text('Results')
    images = fields.One2Many('ir.attachment', 'resource', 'Images')

    # Mostly used in report template.
    def has_image_comments(self):
        return (True in [img.description != '' and
                         img.description != 'From GNU Health camera' and
                         img.description is not None for img in self.images])

    diagnosis = fields.Text('Diagnosis')
    critearea = fields.One2Many(
        'gnuhealth.lab.test.critearea',
        'gnuhealth_lab_id', 'Analytes criteria')

    # Mostly used in report template.
    def has_critearea_remarks(self):
        return (True in [c.remarks != '' and
                         c.remarks is not None for c in self.critearea])

    date_requested = fields.DateTime(
        'Request Date', required=True)
    date_analysis = fields.DateTime('Analysis Date')
    request_order = fields.Integer('Order', readonly=True)

    pathology = fields.Many2One(
        'gnuhealth.pathology', 'Pathology',
        help='Pathology confirmed / associated to this lab test.')

    analytes_summary = fields.Function(
        fields.Text('Summary'), 'get_analytes_summary')

    # From crypto
    state = fields.Selection([
        ('draft', 'Draft'),
        ('done', 'Done'),
        ('validated', 'Validated'),
    ], 'State', readonly=True, sort=False)

    done_by = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Done by', readonly=True, help='Professional who processes this'
        ' lab test',
        states=STATES)

    done_date = fields.DateTime(
        'Finished on', readonly=True,
        states=STATES)

    validated_by = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Validated by', readonly=True, help='Professional who validates this'
        ' lab test',
        states=STATES)

    validation_date = fields.DateTime(
        'Validated on', readonly=True,
        states=STATES)

    historize = fields.Boolean(
        "Historize",
        states=STATES,
        depends=['pathology'],
        help='If this flag is set'
        ' the a new health condition will be added'
        ' to the patient history.'
        ' Unset it if this lab test is in the context'
        ' of a pre-existing condition of the patient.'
        ' The condition will be created when the lab test'
        ' is confirmed and validated')

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_historize():
        return False

    @staticmethod
    def default_date_requested():
        return datetime.now()

    @staticmethod
    def default_date_analysis():
        return datetime.now()

    @staticmethod
    def default_source_type():
        return 'patient'

    @fields.depends('pathology')
    def on_change_with_historize(self):
        if (self.pathology):
            return True

    def get_analytes_summary(self, name):
        summ = ""
        for analyte in self.critearea:
            if analyte.result or analyte.result_text:
                res = ""
                res_text = ""
                if analyte.result_text:
                    res_text = analyte.result_text
                if analyte.result:
                    unit = (analyte.units
                            and analyte.units.name
                            or '')
                    res = str(analyte.result) + f' ({unit})  '
                summ = summ + analyte.rec_name + "  " + \
                    res + res_text + "\n"
        return summ

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'lab_test_sequence', **pattern)
        if sequence:
            return sequence.get()

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('name'):
                values['name'] = cls.generate_code()

        return super(Lab, cls).create(vlist)

    @classmethod
    def search_rec_name(cls, name, clause):
        if clause[1].startswith('!') or clause[1].startswith('not '):
            bool_op = 'AND'
        else:
            bool_op = 'OR'
        return [
            bool_op,
            ('patient', ) + tuple(clause[1:]),
            ('name', ) + tuple(clause[1:]),
        ]

    @classmethod
    @ModelView.button
    def complete_criteareas(cls, labs):
        pool = Pool()
        Critearea = pool.get('gnuhealth.lab.test.critearea')

        lab = labs[0]
        test_cases = []

        for critearea in (lab and lab.test and lab.test.critearea):
            if not Critearea.search(
                    [('gnuhealth_lab_id', '=', lab.id),
                     # NOTE: We do not have 'code' field before, so we
                     # search 'name' field as fallback in this place,
                     # but 'name' field will be translated to other
                     # languages, as search key it is not reliable as
                     # 'code', so we suggest user maintain critearea's
                     # code.
                     ['OR', [('code', '=', critearea.code),
                             ('name', '=', critearea.name)]]]):

                test_cases.append({
                    'gnuhealth_lab_id': lab.id,
                    'name': critearea.name,
                    'code': critearea.code,
                    'test_method': critearea.test_method,
                    'sequence': critearea.sequence,
                    'limits_verified': critearea.limits_verified,
                    'lower_limit': critearea.lower_limit,
                    'upper_limit': critearea.upper_limit,
                    'normal_range': critearea.normal_range,
                    "to_integer": critearea.to_integer,
                    'units': critearea.units and critearea.units.id})

        if test_cases:
            Critearea.create(test_cases)

    @classmethod
    @ModelView.button
    def generate_document(cls, documents):
        document = documents[0]

        # Set the document to "Done"
        # and write the name of the signing health professional

        hp = get_health_professional()

        cls.write(documents, {
            'done_by': hp,
            'done_date': datetime.now(),
            'state': 'done', })

        # Create lab PoL if the person has a federation account.
        if (document.patient and document.patient.party.federation_account):
            cls.create_lab_pol(document)

        # Create Health condition to the patient
        # if there is a confirmed pathology associated and
        # validated to the lab test result
        # The flag historize must also be set
        if (document.pathology and document.historize):
            cls.create_health_condition(document)

    @classmethod
    @ModelView.button
    def set_to_draft(cls, documents):
        cls.write(documents, {
            'state': 'draft', })

    def is_patient(self):
        return (self.source_type == 'patient')

    def is_other_source(self):
        return (self.source_type == 'other_source')

    def find_images(self, critearea_code):
        pool = Pool()
        Attachment = pool.get('ir.attachment')

        images = None
        if critearea_code:
            # We will search images which description include string:
            # '<<critearea_code>>'.
            search_str = '%<<' + critearea_code + '>>%'
            images = Attachment.search(
                [('resource', '=', self),
                 ('description', 'like', search_str)])

        return images

    @classmethod
    def create_health_condition(cls, lab_info):
        """ Create the health condition when specified and
            validated in the lab test
        """
        HealthCondition = Pool().get('gnuhealth.patient.disease')
        health_condition = []

        vals = {
            'patient': lab_info.patient.id,
            'pathology': lab_info.pathology,
            'diagnosed_date': lab_info.date_analysis.date(),
            'lab_confirmed': True,
            'lab_test': lab_info.id,
            'extra_info': lab_info.diagnosis,
            'healthprof': lab_info.requestor
        }

        health_condition.append(vals)
        HealthCondition.create(health_condition)

    @classmethod
    def create_lab_pol(cls, lab_info):
        """ Adds an entry in the person Page of Life
            related to this person lab
        """
        if lab_info.is_patient():
            Pol = Pool().get('gnuhealth.pol')
            pol = []

            test_lines = ""
            for line in lab_info.critearea:
                test_lines = test_lines + line.rec_name + "\n"

            vals = {
                'page': str(uuid4()),
                'person': lab_info.patient.party.id,
                'page_date': lab_info.date_analysis,
                'federation_account':
                    lab_info.patient.party.federation_account,
                'page_type': 'medical',
                'medical_context': 'lab',
                'relevance': 'important',
                'info': lab_info.analytes_summary,
                'author': lab_info.requestor and
                    lab_info.requestor.rec_name
            }

            pol.append(vals)
            Pol.create(pol)

    @classmethod
    def __setup__(cls):
        super(Lab, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('id_uniq', Unique(t, t.name),
             'The test ID code must be unique')
        ]
        cls._order.insert(0, ('date_requested', 'DESC'))
        cls._buttons.update({
            'complete_criteareas': {},
            'generate_document': {
                'invisible': Not(Equal(Eval('state'), 'draft')),
            },
            'set_to_draft': {
                'invisible': Not(Equal(Eval('state'), 'done')),
            },
        })

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None


class GnuHealthLabTestUnits(ModelSQL, ModelView):
    'Lab Test Units'
    __name__ = 'gnuhealth.lab.test.units'

    name = fields.Char(
        'Unit', translate=True)

    code = fields.Char(
        'Code', translate=False)

    @classmethod
    def __setup__(cls):
        super(GnuHealthLabTestUnits, cls).__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            ('name_uniq', Unique(t, t.name),
             'The Unit name must be unique')
        ]

    @classmethod
    def check_xml_record(cls, records, values):
        return True


class GnuHealthTestCritearea(ModelSQL, ModelView):
    'Lab Test Critearea'
    __name__ = 'gnuhealth.lab.test.critearea'

    name = fields.Char(
        'Analyte', required=True, translate=True)

    test_method = fields.Char(
        'Method',
        help='Test method, for example: Real-time PCR ...',
        translate=True)

    excluded = fields.Boolean(
        'Excluded', help='Select this option when'
        ' this analyte is excluded from the test')
    result = fields.Float('Value')

    to_integer = fields.Boolean(
        'To integer',
        help='Convert result value to interger in report.')

    result_text = fields.Text(
        'Qualitative',
        help='Non-numeric results. For '
        'example qualitative values, morphological, colors ...')
    remarks = fields.Text('Remarks')
    normal_range = fields.Text('Reference', translate=True)
    lower_limit = fields.Float('Lower Limit')
    upper_limit = fields.Float('Upper Limit')
    limits_verified = fields.Boolean(
        'Limits verified',
        help='The upper and lower limits have been verified again, '
        'sometimes limits will depend on other indicators of the patient, '
        'such as: age, pregnancy status, etc, so it is very important to '
        'verify them, because warning status depend on limits values.')
    warning = fields.Boolean(
        'Warn', help='Warns the patient about this '
        ' analyte result'
        ' It is useful to contextualize the result to each patient status '
        ' like age, sex, comorbidities, ...')
    units = fields.Many2One('gnuhealth.lab.test.units', 'Units')
    test_type_id = fields.Many2One(
        'gnuhealth.lab.test_type', 'Test type')
    gnuhealth_lab_id = fields.Many2One(
        'gnuhealth.lab', 'Test Cases')
    sequence = fields.Integer('Sequence')

    # code field is mainly used by interface script, for example:
    # gnuhealth_csv_lab_interface.py in example directory.
    ##
    # sequence field is not suitable for interface script, for it may
    # be changed by user for sort reason, when it changed, interface
    # script can not find error. for example: when a criterea
    # sequence is changed from 1 to 2 for sort reason. if interface
    # script do not update, it will run no error and push wrong
    # value.
    ##
    # name field is not suitable for interface stript too, for it
    # will be changed when user use different languages.
    code = fields.Char(
        'Code', translate=False,
        help="Lab test critearea code, "
        "mainly used by lab interface script.")

    # Show the warning icon if warning is active on the analyte line
    lab_warning_icon = fields.Function(fields.Char(
        'Lab Warning Icon'),
        'get_lab_warning_icon')

    def get_lab_warning_icon(self, name):
        if (self.warning):
            return 'gnuhealth-warning'

    # Use by template
    def get_report_warn_indicator(self):
        if (self.result is not None and
            self.lower_limit is not None and
                self.result < self.lower_limit):
            indicator = '↓'
        elif (self.result is not None and
              self.upper_limit is not None and
              self.result > self.upper_limit):
            indicator = '↑'
        elif self.warning:
            indicator = '*'
        else:
            indicator = ' '

        return indicator

    # Use by template
    def get_report_result(self, unit=True, normal_range=True):
        if (self.result is not None):
            if self.to_integer:
                result = str(int(self.result))
            else:
                result = str(self.result)
        else:
            result = ''

        if (self.result is not None) and unit and self.units:
            unit = " " + self.units.name
        else:
            unit = ''

        if (self.result is not None) and normal_range and self.normal_range:
            normal_range = " (" + self.normal_range + ")"
        else:
            normal_range = ''

        if (self.result is not None) and self.result_text:
            result_text = '\n{' + self.result_text + '}'
        elif self.result_text:
            result_text = self.result_text
        else:
            result_text = ''

        return result + unit + normal_range + result_text

    @classmethod
    def __setup__(cls):
        super(GnuHealthTestCritearea, cls).__setup__()
        cls._order.insert(0, ('sequence', 'ASC'))

    @staticmethod
    def default_to_integer():
        return False

    @staticmethod
    def default_sequence():
        return 1

    @staticmethod
    def default_excluded():
        return False

    @staticmethod
    def default_limits_verified():
        return True

    @fields.depends('result', 'warning')
    def on_change_with_lab_warning_icon(self):
        if self.warning:
            return 'gnuhealth-warning'

    @fields.depends('result', 'lower_limit', 'upper_limit')
    def on_change_with_warning(self):
        normal = True

        # Note: do not use 'if (self.result)' code style in here, for
        # in python: 0.0 = False

        # lower_limit < x < upper_limit
        if (self.result is not None
            and self.lower_limit is not None
                and self.upper_limit is not None):
            normal = (self.lower_limit < self.result < self.upper_limit)
        # lower_limit < x, At least lower_limit
        elif (self.result is not None
              and self.lower_limit is not None
              and self.upper_limit is None):
            normal = (self.lower_limit < self.result)
        # x < upper_limit, Up to upper_limit
        elif (self.result is not None
              and self.lower_limit is None
              and self.upper_limit is not None):
            normal = (self.result < self.upper_limit)
        else:
            normal = True

        return (not normal)

    @classmethod
    def check_xml_record(cls, records, values):
        return True

    @classmethod
    def view_attributes(cls):
        return super().view_attributes() + [
            ('/tree', 'visual',
                If(Eval('warning'), 'warning', '')),
        ]


class GnuHealthPatientLabTest(ModelSQL, ModelView):
    'Lab Test Request'
    __name__ = 'gnuhealth.patient.lab.test'

    test_type = fields.Many2One(
        'gnuhealth.lab.test_type', 'Test Type',
        required=True)
    date = fields.DateTime('Date')
    state = fields.Selection([
        ('draft', 'Draft'),
        ('tested', 'Tested'),
        ('ordered', 'Ordered'),
        ('cancel', 'Cancel'),
    ], 'State', readonly=True)
    source_type = fields.Selection([
        ('patient', 'Patient'),
        ('other_source', 'Other')
    ], 'Source',
        help='Sample source type.',
        sort=False)

    patient_id = fields.Many2One(
        'gnuhealth.patient', 'Patient',
        states={'invisible': (Eval('source_type') != 'patient')})

    other_source = fields.Char(
        'Other',
        states={
            'invisible': (
                Eval('source_type') != 'other_source')},
        help="Other sample source.")
    source_name = fields.Function(
        fields.Text('Source name'), 'get_source_name')

    gender_str = fields.Char(
        'Gender', readonly=True,
        states={'invisible': (Eval('source_type') != 'patient')})

    age_num = fields.Float(
        'Age Num',
        digits=(3, 3), readonly=True,
        help='Age year number, '
        '(years x 365 + months x 30.5 + days) / 365',
        states={'invisible': (Eval('source_type') != 'patient')})

    specimen_type = fields.Char(
        'Specimen',
        help='Specimen type, for example: '
        'whole blood, plasma, urine or feces ...')

    doctor_id = fields.Many2One(
        'gnuhealth.healthprofessional', 'Health Prof',
        help="Health professional who requests the lab test.")
    context = fields.Many2One(
        'gnuhealth.pathology', 'Context',
        help="Health context for this order. It can be a suspected or"
             " existing health condition, a regular health checkup, ...")

    comment = fields.Text('Additional Information')
    request = fields.Integer('Order', readonly=True)
    urgent = fields.Boolean('Urgent')

    def get_source_name(self, name):
        if self.is_patient():
            return self.patient_id and self.patient_id.rec_name or ''
        else:
            return (self.other_source or '')

    @staticmethod
    def default_date():
        return datetime.now()

    @staticmethod
    def default_source_type():
        return 'patient'

    @staticmethod
    def default_state():
        return 'draft'

    @staticmethod
    def default_doctor_id():
        return get_health_professional()

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'lab_request_sequence', **pattern)
        if sequence:
            return sequence.get()

    # Update age_num and gender_str based on the patient_id
    @fields.depends(
        'patient_id', '_parent_patient_id.age')
    def on_change_patient_id(self):
        if (self.patient_id):
            self.gender_str = self.patient_id.gender_str
            self.age_num = get_age_for_comparison(
                self.patient_id.age, type='y')
        else:
            self.gender_str = None
            self.age_num = None

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('request'):
                values['request'] = cls.generate_code()

        return super(GnuHealthPatientLabTest, cls).create(vlist)

    @classmethod
    def copy(cls, tests, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['request'] = None
        default['date'] = cls.default_date()
        return super(GnuHealthPatientLabTest, cls).copy(
            tests, default=default)

    def is_patient(self):
        return (self.source_type == 'patient')

    def is_other_source(self):
        return (self.source_type == 'other_source')

    @classmethod
    def __setup__(cls):
        super(GnuHealthPatientLabTest, cls).__setup__()
        cls._order.insert(0, ('date', 'DESC'))
        cls._order.insert(1, ('request', 'DESC'))
        cls._order.insert(2, ('test_type', 'ASC'))

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to test_type
        if (table_h.column_exist('name')
                and not table_h.column_exist('test_type')):
            table_h.column_rename('name', 'test_type')

        super().__register__(module)
        table_h = cls.__table_handler__(module)


class PatientHealthCondition(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient.disease'

    # Adds lab confirmed and the link to the test to the
    # Patient health Condition

    lab_confirmed = fields.Boolean(
        'Lab Confirmed', help='Confirmed by'
        ' laboratory test')

    lab_test = fields.Many2One(
        'gnuhealth.lab', 'Lab Test',
        domain=[('patient', '=', Eval('patient'))], depends=['patient'],
        states={'invisible': Not(Bool(Eval('lab_confirmed')))},
        help='Lab test that confirmed the condition')



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.model.exceptions import ValidationError


class StockMoveExists(ValidationError):
    pass



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                           HEALTH package                              #
#                __init__.py: Wizard declaration file                   #
#########################################################################

from . import wizard_appointment_evaluation
from . import wizard_check_immunization_status
from . import wizard_update_patient_disease_info


__all__ = ['wizard_appointment_evaluation',
           'wizard_check_immunization_status',
           'wizard_update_patient_disease_info']



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH ARCHIVES package                         #
#                   health_archives.py: main module                     #
#########################################################################
from trytond.model import ModelView, ModelSQL, fields, Unique


__all__ = ['PaperArchive']


class PaperArchive(ModelSQL, ModelView):
    'Location of PAPER Patient Clinical History'

    __name__ = 'gnuhealth.paper_archive'

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient', required=True,
        help="Patient associated to this newborn baby")

    legacy = fields.Char(
        'Legacy Code', help="If existing, please enter"
        " the old / legacy code associated to this Clinical History")
    location = fields.Many2One(
        'gnuhealth.hospital.unit', 'Unit', required=True,
        help="Location / Unit where this clinical history document"
        " should reside.")

    hc_status = fields.Selection((
        ('archived', 'Archived'),
        ('borrowed', 'Borrowed'),
        ('lost', 'Lost'),
    ), 'Status', required=True, sort=False)

    current_location = fields.Many2One(
        'gnuhealth.hospital.unit', 'Current Location',
        help="Location / Unit where this clinical history document"
        " should reside.")

    identification_code = fields.Function(
        fields.Char('Code'),
        'get_patient_history', searcher='search_patient_code')

    requested_by = fields.Many2One(
        'party.party', 'Requested by',
        domain=[('is_person', '=', True)],
        help="Person who last requested the document")

    request_date = fields.DateTime("Request Date")
    return_date = fields.DateTime("Returned Date")
    comments = fields.Char("Comments")

    @classmethod
    def __setup__(cls):
        '''Create constraints for both the legacy number and patient'''
        super(PaperArchive, cls).__setup__()

        t = cls.__table__()
        cls._sql_constraints = [
            ('legacy_unique', Unique(t, t.legacy),
                'The history already exists'),
            ('patient_unique', Unique(t, t.patient),
                'The patient history already exists'),

        ]

    @classmethod
    def search_patient_code(cls, name, clause):
        '''Retrieve the Patient Identification Code'''
        res = []
        value = clause[2]
        res.append(('patient.puid', clause[1], value))
        return res

    def get_patient_history(self, name):
        return self.patient.puid



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                       HEALTH NURSING package                          #
#              __init__.py: Package declaration file                    #
#########################################################################

from trytond.pool import Pool
from . import health_nursing
from . import sequences
from . import health


def register():
    Pool.register(
        health.PatientProcedure,
        health_nursing.PatientAmbulatoryCare,
        health_nursing.AmbulatoryCareProcedure,
        sequences.GnuHealthSequences,
        sequences.AmbulatoryCareSequence,
        module='health_nursing', type_='model')



# Copyright (C) 2008-2025 Luis Falcon <falcon@gnuhealth.org>
# Copyright (C) 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections import defaultdict
from sql.aggregate import Count
from sql.functions import DateTrunc
from datetime import date, datetime
from trytond.report import Report
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.modules.health.core import parse_compute_age
from dateutil.relativedelta import relativedelta

from matplotlib import pyplot as plt
from matplotlib.ticker import MaxNLocator
import matplotlib as mpl

from trytond.modules.health.core import (convert_date_timezone,
                                         matplotlib_setup)
from trytond.i18n import gettext

import io

__all__ = ['InstitutionEpidemicsReport']


class InstitutionEpidemicsReport(Report):
    __name__ = 'gnuhealth.epidemics.report'

    @classmethod
    def get_population_with_no_dob(cls):
        """ Return Total Number of living people in the system
        without a date of birth"""
        pool = Pool()
        Party = pool.get('party.party')

        return Party.search([
            ('is_person', '=', True),
            ('deceased', '!=', True),
            ('dob', '=', None),
        ], count=True)

    @classmethod
    def get_population(cls, date1, date2, gender, total):
        """ Return Total Number of living people in the system
        segmented by age group and gender"""
        pool = Pool()
        Party = pool.get('party.party')

        domain = [
            ('deceased', '!=', True),
            ('gender', '=', gender),
        ]

        if not total:
            domain.append(('dob', '>=', date2))
            domain.append(('dob', '<=', date1))

        return Party.search(domain, count=True)

    @classmethod
    def get_new_people(cls, start_date, end_date, in_health_system):
        """ Return Total Number of new registered persons alive """
        pool = Pool()
        Party = pool.get('party.party')

        domain = [
            ('activation_date', '>=', start_date),
            ('activation_date', '<=', end_date),
            ('deceased', '!=', True),
            ('is_person', '=', True),
        ]

        if in_health_system:
            domain.append(('is_patient', '=', True))

        return Party.search(domain, count=True)

    @classmethod
    def get_new_births(cls, start_date, end_date):
        """ Return birth certificates within that period """
        pool = Pool()
        BirthCertificate = pool.get('gnuhealth.birth_certificate')

        return BirthCertificate.search([
            ('dob', '>=', start_date),
            ('dob', '<=', end_date),
        ], count=True)

    @classmethod
    def get_new_deaths(cls, start_date, end_date):
        """ Return death certificates within that period """
        """ Truncate the timestamp of DoD to match a whole day"""
        pool = Pool()
        DeathCertificate = pool.get('gnuhealth.death_certificate')
        table = DeathCertificate.__table__()

        dod = DateTrunc('day', table.dod)

        cursor = Transaction().connection.cursor()
        cursor.execute(*table.select(
            Count(table.dod),
            where=((dod >= start_date) & (dod <= end_date))))
        return cursor.fetchone()

    @classmethod
    def get_confirmed_cases(cls, start_date, end_date, dx):
        """ Return number of confirmed cases """

        Condition = Pool().get('gnuhealth.patient.disease')

        clause = [
            ('diagnosed_date', '>=', start_date),
            ('diagnosed_date', '<=', end_date),
        ]

        if dx:
            clause.append(('pathology', '=', dx))

        res = Condition.search(clause)

        return (res)

    @classmethod
    def get_epi_by_day(cls, start_date, end_date, dx):
        """ Return number of confirmed cases """

        Condition = Pool().get('gnuhealth.patient.disease')

        current_day = start_date
        aggr = []
        while current_day <= end_date:
            current_day = current_day + relativedelta(days=1)

            clause = [
                ('diagnosed_date', '=', current_day),
            ]

            if dx:
                clause.append(('pathology', '=', dx))

            res = Condition.search(clause)
            cases_day = len(res)
            daily_data = {'date': current_day, 'cases': cases_day}
            aggr.append(daily_data)
        return (aggr)

    # Death Certificates by day
    @classmethod
    def get_deaths_by_day(cls, start_date, end_date, dx):
        """ Return number of death related to the condition
            Includes both the ultimate case as well as those
            certificates that have the condition as a leading cause
        """

        DeathCert = Pool().get('gnuhealth.death_certificate')

        current_day = start_date
        aggr = []
        while current_day <= end_date:
            cur_day_time = datetime.combine((current_day), datetime.min.time())

            utc_from = convert_date_timezone(cur_day_time, 'utc')
            utc_to = convert_date_timezone(cur_day_time +
                                           relativedelta(days=1), 'utc')

            clause = [
                ('dod', '>=', utc_from),
                ('dod', '<', utc_to)
            ]

            res = DeathCert.search(clause)

            # Reset the cases for each day
            as_immediate_cause = 0
            as_underlying_condition = 0

            # Get the immediate cause of death and the underlying conditions
            # for each certificate on each day.

            for cert in res:
                if (cert.cod.id == dx):
                    as_immediate_cause = as_immediate_cause + 1
                for underlying_condition in cert.underlying_conditions:
                    if (underlying_condition.condition.id == dx):
                        as_underlying_condition = as_underlying_condition + 1

            daily_data = {'date': current_day,
                          'certs_day_ic': as_immediate_cause,
                          'certs_day_uc': as_underlying_condition}
            aggr.append(daily_data)
            current_day = current_day + relativedelta(days=1)

        return (aggr)

    @classmethod
    def plot_cases_timeseries(cls, start_date, end_date,
                              health_condition_id, hc):

        epi_series = cls.get_epi_by_day(start_date,
                                        end_date, health_condition_id)

        days = []
        cases_day = []
        for day in epi_series:
            days.append(day['date'])
            # Confirmed cases by day
            cases_day.append(day['cases'])

        fig = plt.figure(figsize=(6, 3))
        cases_by_day = fig.add_subplot(1, 1, 1)
        cases_by_day.bar(days, cases_day)
        cases_by_day.yaxis.set_major_locator(MaxNLocator(integer=True))
        fig.autofmt_xdate()

        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        image = holder.getvalue()

        holder.close()
        return (image)

    @classmethod
    def plot_deaths_timeseries(cls, start_date,
                               end_date, health_condition_id, hc):

        death_certs = cls.get_deaths_by_day(start_date,
                                            end_date, health_condition_id)

        days = []
        certs_ic_day = []
        certs_uc_day = []
        for day in death_certs:
            days.append(day['date'])
            # Death certificates as an immediate cause
            certs_ic_day.append(day['certs_day_ic'])
            # Death certificates as an underlying cause
            certs_uc_day.append(day['certs_day_uc'])

        fig = plt.figure(figsize=(6, 3))
        deaths_by_day = fig.add_subplot(1, 1, 1)
        deaths_by_day.plot(
            days, certs_ic_day, label=gettext(
                "health_reporting.msg_plot_label_immediate_cause_str"))
        deaths_by_day.plot(
            days, certs_uc_day,
            label=gettext(
                "health_reporting.msg_plot_label_underlying_condition_str"))
        deaths_by_day.yaxis.set_major_locator(MaxNLocator(integer=True))
        deaths_by_day.legend()

        fig.autofmt_xdate()

        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        image = holder.getvalue()

        holder.close()
        return (image)

    @classmethod
    def plot_cases_ethnicity(cls, start_date, end_date, ethnic_count, hc):

        for k, v in list(ethnic_count.items()):
            if (v == 0):
                # Remove ethnicities with zero cases from the plot
                del (ethnic_count[k])

        fig = plt.figure(figsize=(6, 3))
        cases_by_ethnicity = fig.add_subplot(1, 1, 1)
        cases_by_ethnicity.pie(ethnic_count.values(),
                               autopct='%1.1f%%',
                               labels=ethnic_count.keys())

        fig.autofmt_xdate()

        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        image = holder.getvalue()

        holder.close()
        return (image)

    @classmethod
    def get_ethnic_groups(cls):
        # Build a list with the ethnic groups
        Condition = Pool().get('gnuhealth.ethnicity')
        ethnic_groups = Condition.search([])
        ethnicities = []
        for ethnic_group in ethnic_groups:
            ethnicities.append(ethnic_group.name)
        return (ethnicities)

    @classmethod
    def plot_cases_socioeconomics(cls, start_date, end_date, ses_count, hc):

        for k, v in list(ses_count.items()):
            if (v == 0):
                # Remove socioeconomic groups with zero cases from the plot
                del (ses_count[k])

        fig = plt.figure(figsize=(6, 3))
        cases_by_socioeconomics = fig.add_subplot(1, 1, 1)
        cases_by_socioeconomics.pie(ses_count.values(),
                                    autopct='%1.1f%%',
                                    labels=ses_count.keys())

        fig.autofmt_xdate()

        holder = io.BytesIO()
        fig.savefig(holder, format="svg")
        image = holder.getvalue()

        holder.close()
        return (image)

    @classmethod
    def get_context(cls, records, header, data):

        Condition = Pool().get('gnuhealth.pathology')

        ethnic_groups = cls.get_ethnic_groups()

        ethnic_count = defaultdict(int)

        ses_count = defaultdict(int)

        context = super(InstitutionEpidemicsReport, cls).get_context(
            records, header, data)

        start_date = data['start_date']
        context['start_date'] = data['start_date']

        end_date = data['end_date']
        context['end_date'] = data['end_date']

        context['demographics'] = data['demographics']

        health_condition_id = data['health_condition']

        hc = Condition.search(
            [('id', '=', health_condition_id)], limit=1)[0]

        context['health_condition'] = hc

        # Demographics
        today = date.today()

        context[''.join(['p', 'total_', 'f'])] = \
            cls.get_population(None, None, 'f', total=True)

        context[''.join(['p', 'total_', 'm'])] = \
            cls.get_population(None, None, 'm', total=True)

        # Living people with NO date of birth
        context['no_dob'] = \
            cls.get_population_with_no_dob()

        # Build the Population Pyramid for registered people

        for age_group in range(0, 21):
            date1 = today - relativedelta(years=(age_group * 5))
            date2 = today - relativedelta(years=((age_group * 5) + 5), days=-1)

            context[''.join(['p', str(age_group), 'f'])] = \
                cls.get_population(date1, date2, 'f', total=False)
            context[''.join(['p', str(age_group), 'm'])] = \
                cls.get_population(date1, date2, 'm', total=False)

        # Count those lucky over 105 years old :)
        date1 = today - relativedelta(years=105)
        date2 = today - relativedelta(years=200)

        context['over105f'] = \
            cls.get_population(date1, date2, 'f', total=False)
        context['over105m'] = \
            cls.get_population(date1, date2, 'm', total=False)

        # Count registered people, and those within the system of health
        context['new_people'] = \
            cls.get_new_people(start_date, end_date, False)
        context['new_in_health_system'] = \
            cls.get_new_people(start_date, end_date, in_health_system=True)

        # New births
        context['new_births'] = \
            cls.get_new_births(start_date, end_date)

        # New deaths
        context['new_deaths'] = \
            cls.get_new_deaths(start_date, end_date)

        # Get cases within the specified date range

        confirmed_cases = cls.get_confirmed_cases(start_date, end_date,
                                                  health_condition_id)

        context['confirmed_cases'] = confirmed_cases

        epidemics_dx = []
        non_age_cases = 0
        cases_f = 0
        cases_m = 0

        # Global Condition info
        for confirmed_case in confirmed_cases:
            # Sex distribution
            if (confirmed_case.patient.gender == 'f'):
                cases_f += 1
            else:
                cases_m += 1

            # Ethnic groups distribution
            if (confirmed_case.patient.party.ethnic_group):
                ethnicity = confirmed_case.patient.party.ethnic_group.name
                if (ethnicity in ethnic_groups):
                    ethnic_count[ethnicity] = ethnic_count[ethnicity] + 1

            # Socioeconomic groups distribution
            if (confirmed_case.patient.ses):
                ses_str = confirmed_case.patient.ses_str
                ses_count[ses_str] += 1

            if not confirmed_case.patient.age:
                non_age_cases += 1

        total_cases = len(confirmed_cases)

        context['confirmed_cases_num'] = total_cases
        context['cases_f'] = cases_f
        context['cases_m'] = cases_m
        context['non_age_cases'] = non_age_cases

        group_1 = group_2 = group_3 = group_4 = group_5 = 0
        group_1f = group_2f = group_3f = group_4f = group_5f = 0

        for case in confirmed_cases:

            if (case.patient.age):

                # Strip to get the raw year
                age_year = parse_compute_age(case.patient.age)[0]

                # Age groups in this diagnostic
                if (age_year < 5):
                    group_1 += 1
                    if (case.patient.gender == 'f'):
                        group_1f += 1
                if (age_year in range(5, 14)):
                    group_2 += 1
                    if (case.patient.gender == 'f'):
                        group_2f += 1
                if (age_year in range(15, 45)):
                    group_3 += 1
                    if (case.patient.gender == 'f'):
                        group_3f += 1
                if (age_year in range(46, 60)):
                    group_4 += 1
                    if (case.patient.gender == 'f'):
                        group_4f += 1
                if (age_year > 60):
                    group_5 += 1
                    if (case.patient.gender == 'f'):
                        group_5f += 1

        cases = {'diagnosis': health_condition_id,
                 'age_group_1': group_1, 'age_group_1f': group_1f,
                 'age_group_2': group_2, 'age_group_2f': group_2f,
                 'age_group_3': group_3, 'age_group_3f': group_3f,
                 'age_group_4': group_4, 'age_group_4f': group_4f,
                 'age_group_5': group_5, 'age_group_5f': group_5f,
                 'total': total_cases, }

        # Append into the report list the resulting
        # dictionary entry

        epidemics_dx.append(cases)

        context['epidemics_dx'] = epidemics_dx

        # Configure matplotlib, for example: font.
        matplotlib_setup(mpl)

        # New cases by day
        context['cases_timeseries'] = cls.plot_cases_timeseries(
            start_date,
            end_date,
            health_condition_id, hc)

        # Cases by ethnic groups
        context['cases_ethnicity'] = cls.plot_cases_ethnicity(
            start_date,
            end_date, ethnic_count, hc)

        # Cases by Socioeconomic groups
        context['cases_ses'] = cls.plot_cases_socioeconomics(
            start_date,
            end_date, ses_count, hc)

        # Death certificates by day
        context['deaths_timeseries'] = cls.plot_deaths_timeseries(
            start_date,
            end_date,
            health_condition_id, hc)

        return context



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
from . import wizard_health_pediatrics_growth_charts_who


__all__ = ['wizard_health_pediatrics_growth_charts_who']



# Copyright (C) 2008-2025 Luis Falcon <lfalcon@gnusolidario.org>
# Copyright (C) 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import health_disability
from trytond.pool import Pool


def register():
    Pool.register(
        health_disability.GnuHealthPatient,
        health_disability.Product,
        health_disability.BodyFunctionCategory,
        health_disability.BodyFunction,
        health_disability.BodyStructureCategory,
        health_disability.BodyStructure,
        health_disability.ActivityAndParticipationCategory,
        health_disability.ActivityAndParticipation,
        health_disability.EnvironmentalFactorCategory,
        health_disability.EnvironmentalFactor,
        health_disability.PatientDisabilityAssessment,
        health_disability.PatientBodyFunctionAssessment,
        health_disability.PatientBodyStructureAssessment,
        health_disability.PatientActivityAndParticipationAsssessment,
        health_disability.PatientEnvironmentalFactorAssessment,
        health_disability.PatientAmputation,
        health_disability.PatientProthesis,
        health_disability.PatientData,
        module='health_disability', type_='model')



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                     HEALTH NURSING package                            #
#                health_nursing.py: main module                         #
#########################################################################
from trytond.model import ModelView, ModelSQL, fields
from datetime import datetime
from trytond.pool import Pool
from trytond.pyson import Eval, Equal
from trytond.i18n import gettext
from trytond.modules.health.core import get_health_professional

from .exceptions import (
    NoAssociatedHealthProfessional
)

__all__ = [
    'PatientAmbulatoryCare', 'AmbulatoryCareProcedure']


class PatientAmbulatoryCare(ModelSQL, ModelView):
    'Ambulatory Care'
    __name__ = 'gnuhealth.patient.ambulatory_care'

    STATES = {'readonly': Eval('state') == 'done'}

    name = fields.Char('ID', readonly=True)
    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient',
        required=True, states=STATES)

    state = fields.Selection([
        (None, ''),
        ('draft', 'In Progress'),
        ('done', 'Done'),
    ], 'State', readonly=True)

    base_condition = fields.Many2One(
        'gnuhealth.pathology', 'Condition',
        states=STATES)
    evaluation = fields.Many2One(
        'gnuhealth.patient.evaluation',
        'Related Evaluation', domain=[('patient', '=', Eval('patient'))],
        depends=['patient'], states=STATES)
    ordering_professional = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Requested by', states=STATES)
    health_professional = fields.Many2One(
        'gnuhealth.healthprofessional',
        'Health Prof', readonly=True)

    ambulatory_procedures = fields.One2Many(
        'gnuhealth.patient.procedure', 'reference', 'Procedures',
        domain=[
            ('patient', '=', Eval('patient')),
            ('ctx', '=', 'ambulatory'),
            # ('pdate', '=', Eval('session_start')),
        ],
        depends=['patient'],
        help='Procedures done during the ambulatory care session')

    # Deprecated in GH 5.0 by ambulatory_procedures
    procedures = fields.One2Many(
        'gnuhealth.ambulatory_care_procedure', 'ambcare',
        'Procedures', states=STATES,
        help="List of the procedures in this session. Please enter the first "
        "one as the main procedure")
    session_number = fields.Integer('Session #', states=STATES)
    session_start = fields.DateTime('Start', required=True, states=STATES)

    # Vital Signs
    systolic = fields.Integer('Systolic Pressure', states=STATES)
    diastolic = fields.Integer('Diastolic Pressure', states=STATES)
    bpm = fields.Integer(
        'Heart Rate', states=STATES,
        help='Heart rate expressed in beats per minute')
    respiratory_rate = fields.Integer(
        'Respiratory Rate', states=STATES,
        help='Respiratory rate expressed in breaths per minute')
    osat = fields.Integer(
        'Oxygen Saturation', states=STATES,
        help='Oxygen Saturation(arterial).')
    temperature = fields.Float(
        'Temperature', states=STATES,
        help='Temperature in celsius')

    warning = fields.Boolean(
        'Warning', help="Check this box to alert the "
        "supervisor about this session. A warning icon will be shown in the "
        "session list", states=STATES)
    warning_icon = fields.Function(
        fields.Char('Warning Icon'), 'get_warn_icon')

    # Glycemia
    glycemia = fields.Integer(
        'Glycemia', help='Blood Glucose level',
        states=STATES)

    weight = fields.Integer(
        'Weight',
        help="Measured weight, in kg")

    pain = fields.Boolean(
        'Pain',
        help="Check if the patient is in pain")

    pain_level = fields.Integer(
        'Pain level',
        help="Enter the pain level, from 1 to 10.")

    evolution = fields.Selection([
        (None, ''),
        ('initial', 'Initial'),
        ('n', 'Status Quo'),
        ('i', 'Improving'),
        ('w', 'Worsening'),
    ], 'Evolution', help="Check your judgement of current "
        "patient condition", sort=False, states=STATES)

    evolution_str = evolution.translated('evolution')

    session_end = fields.DateTime('End', readonly=True)
    next_session = fields.DateTime('Next Session', states=STATES)
    session_notes = fields.Text('Notes', states=STATES)

    signed_by = fields.Many2One(
        'gnuhealth.healthprofessional', 'Signed by', readonly=True,
        states={'invisible': Equal(Eval('state'), 'draft')},
        help="Health Professional that signed the session")

    @staticmethod
    def default_ambulatory_procedures():
        """When creating a new patient ambulatory session,
           GNU Health checks if there is one code set on the
           gnuhealth.procedures.config model, and use it as the first line
           on the procedure list.
           The user can remove it or use another in that ambulatory care
           context.
        """
        ProceduresConfig = Pool().get('gnuhealth.procedures.config')(1)
        if (ProceduresConfig and ProceduresConfig.ambulatory_care):
            medical_procedure = int(ProceduresConfig.ambulatory_care)

            return [{'procedure': medical_procedure}]
        else:
            return []

    @staticmethod
    def default_health_professional():
        return get_health_professional()

    @staticmethod
    def default_session_start():
        return datetime.now()

    @staticmethod
    def default_state():
        return 'draft'

    def get_report_pain_and_level(self):
        if self.pain and self.pain_level:
            return gettext('health_nursing.msg_report_pain_level',
                           pain_level=str(self.pain_level))
        elif self.pain:
            return gettext('health_nursing.msg_report_pain_yes')
        else:
            return gettext('health_nursing.msg_report_pain_no')

    @classmethod
    def __setup__(cls):
        super(PatientAmbulatoryCare, cls).__setup__()
        cls._buttons.update({
            'end_session': {
                'invisible': ~Eval('state').in_(['draft']),
            }})

        cls._order.insert(0, ('session_start', 'DESC'))

        # Do not cache default_key as it depends on time
        cls.__rpc__['default_get'].cache = None

    @classmethod
    @ModelView.button
    def end_session(cls, sessions):
        # End the session and discharge the patient
        # Change the state of the session to "Done"
        signing_hp = get_health_professional()

        cls.write(sessions, {
            'state': 'done',
            'signed_by': signing_hp,
            'session_end': datetime.now()
        })

    @classmethod
    def validate(cls, records):
        super(PatientAmbulatoryCare, cls).validate(records)
        for record in records:
            record.check_health_professional()

    def check_health_professional(self):
        if not self.health_professional:
            raise NoAssociatedHealthProfessional(
                gettext('health.msg_no_associated_health_professional'))

    @classmethod
    def generate_code(cls, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'ambulatory_care_sequence', **pattern)
        if sequence:
            return sequence.get()

    @classmethod
    def create(cls, vlist):
        vlist = [x.copy() for x in vlist]
        for values in vlist:
            if not values.get('name'):
                values['name'] = cls.generate_code()
        return super(PatientAmbulatoryCare, cls).create(vlist)

    @classmethod
    def copy(cls, ambulatorycares, default=None):
        if default is None:
            default = {}
        default = default.copy()
        default['name'] = None
        default['session_start'] = cls.default_session_start()
        default['session_end'] = cls.default_session_start()
        return super(
            PatientAmbulatoryCare, cls).copy(
                ambulatorycares,
                default=default)

    def get_warn_icon(self, name):
        if self.warning:
            return 'gnuhealth-warning'


class AmbulatoryCareProcedure(ModelSQL, ModelView):
    'Ambulatory Care Procedure'
    __name__ = 'gnuhealth.ambulatory_care_procedure'

    ambcare = fields.Many2One('gnuhealth.patient.ambulatory_care', 'Session')
    procedure = fields.Many2One(
        'gnuhealth.procedure', 'Code', required=True,
        help="Procedure Code")
    comments = fields.Char('Comments')

    @classmethod
    def __register__(cls, module):
        table_h = cls.__table_handler__(module)

        # Migration from 4.4: rename name to ambcare
        if (table_h.column_exist('name')
                and not table_h.column_exist('ambcare')):
            table_h.column_rename('name', 'ambcare')

        super().__register__(module)
        table_h = cls.__table_handler__(module)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.model import fields
from trytond.pool import PoolMeta

__all__ = ['Invoice', 'InvoiceLine']


class Invoice(metaclass=PoolMeta):
    __name__ = 'account.invoice'

    patient = fields.Function(
        fields.Many2One(
            'gnuhealth.patient', 'Patient',
            help="Patient in the invoice"),
        'get_patient')

    health_service = fields.Function(
        fields.Many2One(
            'gnuhealth.health_service', 'Health Service',
            help="The service entry"),
        'get_health_service', searcher='search_health_service')

    def get_patient(self, name):
        try:
            return self.lines[0].origin.service.patient.id
        except BaseException:
            return None

    def get_health_service(self, name):
        try:
            return self.lines[0].origin.service.id
        except BaseException:
            return None

    @classmethod
    def search_health_service(cls, name, clause):
        try:
            return [
                ('lines.origin.service.id',
                 clause[1],
                 clause[2],
                 'gnuhealth.health_service.line')]
        except BaseException:
            return None


class InvoiceLine(metaclass=PoolMeta):
    __name__ = 'account.invoice.line'

    @classmethod
    def _get_origin(cls):
        return super(InvoiceLine, cls)._get_origin() + [
            'gnuhealth.health_service.line'
        ]



# Copyright (C) 2008-2025 Luis Falcon <lfalcon@gnusolidario.org>
# Copyright (C) 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from trytond.model import ModelView, fields
from trytond.pyson import Eval
from trytond.wizard import Wizard
from trytond.pool import Pool


__all__ = ['RequestPatientImagingTestStart', 'RequestPatientImagingTest']


# Include services in the wizard
class RequestPatientImagingTestStart(ModelView):
    'Request Patient Imaging Test Start'
    __name__ = 'gnuhealth.patient.imaging.test.request.start'

    ungroup_tests = fields.Boolean(
        'Ungroup',
        help="Check if you DO NOT want to include each individual Dx"
        " imaging test from this order in the imaging test generation step."
        " This is useful when some services are not provided in"
        " the same institution.\n"
        "In this case, you need to individually update the service"
        " document from each individual test")

    service = fields.Many2One(
        'gnuhealth.health_service', 'Service',
        domain=[('patient', '=', Eval('patient'))], depends=['patient'],
        help="Service document associated to this Imaging Request")


class RequestPatientImagingTest(Wizard):
    'Request Patient Imaging Test'
    __name__ = 'gnuhealth.patient.imaging.test.request'

    def generate_code(self, **pattern):
        Config = Pool().get('gnuhealth.sequences')
        config = Config(1)
        sequence = config.get_multivalue(
            'imaging_req_seq', **pattern)
        if sequence:
            return sequence.get()

    def append_services(self, imgtest, service):
        """ If the ungroup flag is not set, append the img test
            to the associated health service
        """
        HealthService = Pool().get('gnuhealth.health_service')

        hservice = []

        service_data = {}
        service_lines = []

        # Add the imgtest to the service document

        service_lines.append(('create', [{
            'product': imgtest.product.id,
            'desc': imgtest.product.rec_name,
            'qty': 1
        }]))

        hservice.append(service)
        service_data['service_line'] = service_lines

        HealthService.write(hservice, service_data)

    def transition_request(self):
        ImagingTestRequest = Pool().get('gnuhealth.imaging.test.request')
        request_number = self.generate_code()
        imaging_tests = []
        count = len(self.start.tests)
        num = 1
        for test in self.start.tests:
            imaging_test = {}
            imaging_test['request'] = request_number
            request_line = f'{request_number}-{count:02}-{num:02}'
            imaging_test['request_line'] = request_line
            imaging_test['requested_test'] = test.id
            imaging_test['patient'] = self.start.patient.id
            if self.start.doctor:
                imaging_test['doctor'] = self.start.doctor.id
            if self.start.context:
                imaging_test['context'] = self.start.context.id
            imaging_test['date'] = self.start.date
            imaging_test['urgent'] = self.start.urgent
            if self.start.service:
                imaging_test['service'] = self.start.service.id
                # Append the test directly to the health service document
                # if the Ungroup flag is not set (default).
                if not self.start.ungroup_tests:
                    self.append_services(test, self.start.service)
                    imaging_test['service_updated'] = 'yes'

            imaging_tests.append(imaging_test)
            num = num + 1
        ImagingTestRequest.create(imaging_tests)

        return 'end'



# SPDX-FileCopyrightText: 2012-2017 Cédric Krier
# SPDX-FileCopyrightText: 2017-2024 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2017-2024 Luis Falcon <falcon@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import socketserver
import socket
import http.server
import urllib.parse
import time
import urllib.request
import urllib.parse
import urllib.error
import logging
import os
from threading import local, Thread
import xml.dom.minidom
from base64 import decodebytes
from pywebdav.lib import WebDAVServer, iface
from pywebdav.lib.errors import (
    DAV_Error, DAV_NotFound, DAV_Secret,
    DAV_Forbidden, DAV_Requested_Range_Not_Satisfiable)
from pywebdav.lib.constants import (
    COLLECTION, DAV_VERSION_1, DAV_VERSION_2)
from pywebdav.lib.utils import get_urifilename, quote_uri
from pywebdav.lib.davcmd import (
    copyone, copytree, moveone, movetree, delone, deltree)
from trytond.security import login
from trytond.pool import Pool
from trytond.transaction import Transaction
from trytond.config import config
from trytond.exceptions import (
    UserError, UserWarning, ConcurrencyException)
domimpl = xml.dom.minidom.getDOMImplementation()

DAV_VERSION_1['version'] += ',access-control'
DAV_VERSION_2['version'] += ',access-control'

logger = logging.getLogger(__name__)


def SSLSocket(socket):
    # Let the import error raise only when used
    import ssl
    return ssl.wrap_socket(
        socket,
        server_side=True,
        certfile=config.get('ssl', 'certificate'),
        keyfile=config.get('ssl', 'privatekey'),
        ssl_version=ssl.PROTOCOL_SSLv23)


class Local(local):
    def __init__(self):
        super(Local, self).__init__()
        self.cache = {}


LOCAL = Local()


def setupConfig():

    class ConfigDAV:
        lockemulation = False
        verbose = False
        baseurl = ''

        def getboolean(self, name):
            return bool(self.get(name))

        def get(self, name, default=None):
            return getattr(self, name, default)

    class Config:
        DAV = ConfigDAV()

    return Config()


class BaseThreadedHTTPServer(
        socketserver.ThreadingMixIn,
        http.server.HTTPServer):
    timeout = 1

    def server_bind(self):
        self.socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_REUSEADDR, 1)
        self.socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_KEEPALIVE, 1)
        http.server.HTTPServer.server_bind(self)


class SecureThreadedHTTPServer(BaseThreadedHTTPServer):

    def __init__(self, server_address, HandlerClass):
        BaseThreadedHTTPServer.__init__(self, server_address, HandlerClass)
        self.socket = socket.socket(self.address_family, self.socket_type)
        self.server_bind()
        self.server_activate()


class WebDAVServerThread(Thread):

    def __init__(self, interface, port, secure=False):
        Thread.__init__(self, name='WebDAVServerThread')
        self.secure = secure
        self.ipv6 = False
        for family, _, _, _, _ in socket.getaddrinfo(
                interface or None, port,
                socket.AF_UNSPEC, socket.SOCK_STREAM):
            if family == socket.AF_INET6:
                self.ipv6 = True
            break
        if self.secure:
            handler_class = SecureWebDAVAuthRequestHandler
            server_class = SecureThreadedHTTPServer
            if self.ipv6:
                server_class = SecureThreadedHTTPServer6
        else:
            handler_class = WebDAVAuthRequestHandler
            server_class = BaseThreadedHTTPServer
            if self.ipv6:
                server_class = BaseThreadedHTTPServer6
        handler_class._config = setupConfig()
        handler_class.IFACE_CLASS = TrytonDAVInterface(interface, port, secure)
        handler_class.IFACE_CLASS.baseurl = handler_class._config.DAV.baseurl
        self.server = server_class((interface, port), handler_class)

    def stop(self):
        self.server.shutdown()
        self.server.socket.shutdown(socket.SHUT_RDWR)
        self.server.server_close()
        return

    def run(self):
        self.server.serve_forever()
        return True


class BaseThreadedHTTPServer6(BaseThreadedHTTPServer):
    address_family = socket.AF_INET6


class SecureThreadedHTTPServer6(SecureThreadedHTTPServer):
    address_family = socket.AF_INET6


class TrytonDAVInterface(iface.dav_interface):

    def __init__(self, interface, port, secure=False):
        if secure:
            protocol = 'https'
        else:
            protocol = 'http'
        self.baseuri = '%s://%s:%s/' % (protocol, interface or
                                        socket.gethostname(), port)
        self.verbose = False

    def _log_exception(self, exception):
        if isinstance(exception,
                      (ConcurrencyException, UserError,
                       UserWarning, DAV_Error, DAV_NotFound, DAV_Secret,
                       DAV_Forbidden)):
            logger.debug('Exception %s', exception, exc_info=True)
        else:
            logger.error('Exception %s', exception, exc_info=True)

    @staticmethod
    def get_dburi(uri):
        # URI format :
        # http[s]://<servername>:8080/<dbname>/Calendars/<CalendarName>
        uri = urllib.parse.urlsplit(uri)[2]

        if uri and uri[0] == '/':
            uri = uri[1:]

        dbname, uri = (uri.split('/', 1) + [None])[0:2]
        if dbname:
            dbname = urllib.parse.unquote_plus(dbname)
        if uri:
            uri = urllib.parse.unquote_plus(uri)
        return dbname, uri

    def _get_dburi(self, uri):
        return TrytonDAVInterface.get_dburi(uri)

    def get_childs(self, uri, filter=None):
        res = []
        dbname, dburi = self._get_dburi(uri)
        if not dbname:
            with Transaction().start(None, 0, close=True) as transaction:
                list_ = transaction.database.list()
            for dbname in list_:
                res.append(urllib.parse.urljoin(uri, dbname))
            return res
        pool = Pool(Transaction().database.name)
        try:
            Collection = pool.get('webdav.collection')

            scheme, netloc, path, params, query, fragment = \
                urllib.parse.urlparse(uri)

            if path[-1:] != '/':
                path = path + '/'
            for child in Collection.get_childs(
                    dburi, filter=filter, cache=LOCAL.cache):

                path_child = (path + child)
                res.append(urllib.parse.urlunparse(
                    (scheme, netloc,
                     path_child, params,
                     query, fragment)))
        except KeyError:
            return res
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)

        return res

    def get_data(self, uri, range=None):

        dbname, dburi = self._get_dburi(uri)
        if not dbname or (self.exists(uri) and self.is_collection(uri)):
            res = ('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 '
                   'Transitional//EN">')
            res += '<html>'
            res += '<head>'
            res += ('<meta http-equiv="Content-Type" content="text/html; '
                    'charset=utf-8">')
            res += '<title>Tryton - WebDAV - %s</title>' % dbname or 'root'
            res += '</head>'
            res += '<body>'
            res += '<h2>Collection: %s</h2>' % (get_urifilename(uri) or '/')
            res += '<ul>'
            if dbname:
                scheme, netloc, path, params, query, fragment = \
                    urllib.parse.urlparse(uri)
                if path[-1:] != '/':
                    path += '/'
                res += ('<li><a href="%s">..</a></li>'
                        % urllib.parse.urlunparse(
                            (scheme, netloc, path + '..',
                             params, query, fragment)))
            childs = sorted(self.get_childs(uri))
            for child in childs:
                res += ('<li><a href="%s">%s</a></li>'
                        % (quote_uri(child), get_urifilename(child)))
            res += '</ul>'
            res += '<hr noshade>'
            res += ('<em>Powered by <a href="http://www.tryton.org/">'
                    'GNU Health</a> version %s</em>' %
                    os.environ['GNUHEALTH_VERSION'])
            res += '</body>'
            res += '</html>'
            return res
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')

        try:
            res = Collection.get_data(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        if range is None:
            return res
        size = len(res)
        if range[1] == '':
            range[1] = size
        else:
            range[1] = int(range[1])
        if range[1] > size:
            range[1] = size
        if range[0] == '':
            range[0] = size - range[1]
        else:
            range[0] = int(range[0])
        if range[0] > size:
            raise DAV_Requested_Range_Not_Satisfiable

        return res[range[0]:range[1]]

    def put(self, uri, data, content_type=''):
        res = []
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            raise DAV_Forbidden
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.put(dburi, data, content_type, cache=LOCAL.cache)
            Transaction().commit()
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            Transaction().rollback()
        except Exception as exception:
            self._log_exception(exception)
            Transaction().rollback()
            raise DAV_Error(500)
        if res:
            uparts = list(urllib.parse.urlsplit(uri))
            uparts[2] = res
            res = urllib.parse.urlunsplit(uparts)
        return res

    def mkcol(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            raise DAV_Forbidden
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.mkcol(dburi, cache=LOCAL.cache)
            Transaction().commit()
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            Transaction().rollback()
            raise
        except Exception as exception:
            self._log_exception(exception)
            Transaction().rollback()
            raise DAV_Error(500)
        return res

    def _get_dav_resourcetype(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return COLLECTION
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.get_resourcetype(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return res

    def _get_dav_displayname(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return uri.split('/')[-1]
        pool = Pool(Transaction().database.name)
        try:
            Collection = pool.get('webdav.collection')
            res = Collection.get_displayname(dburi, cache=LOCAL.cache)
        except KeyError:
            raise DAV_NotFound
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return res

    def _get_dav_getcontentlength(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return '0'
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.get_contentlength(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return res

    def _get_dav_getcontenttype(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or self.is_collection(uri):
            return "text/html"
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.get_contenttype(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return res

    def _get_dav_getetag(self, uri):
        return '"' + str(self.get_lastmodified(uri)) + '"'

    def get_creationdate(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return time.time()
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.get_creationdate(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return float(res)

    def get_lastmodified(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return time.time()
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.get_lastmodified(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return float(res)

    def rmcol(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return 403
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.rmcol(dburi, cache=LOCAL.cache)
            Transaction().commit()
        except Exception as exception:
            self._log_exception(exception)
            Transaction().rollback()
            return 500
        return res

    def rm(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return 403
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.rm(dburi, cache=LOCAL.cache)
            Transaction().commit()
        except Exception as exception:
            self._log_exception(exception)
            Transaction().rollback()
            return 500
        return res

    def exists(self, uri):
        dbname, dburi = self._get_dburi(uri)
        if not dbname or not dburi:
            return 1
        pool = Pool(Transaction().database.name)
        Collection = pool.get('webdav.collection')
        try:
            res = Collection.exists(dburi, cache=LOCAL.cache)
        except (DAV_Error, DAV_NotFound,
                DAV_Secret, DAV_Forbidden) as exception:
            self._log_exception(exception)
            raise
        except Exception as exception:
            self._log_exception(exception)
            raise DAV_Error(500)
        return res

    def is_collection(self, uri):
        if self._get_dav_resourcetype(uri) == COLLECTION:
            return 1
        return 0

    def copyone(self, src, dst, overwrite):
        return copyone(self, src, dst, overwrite)

    def copytree(self, src, dst, overwrite):
        return copytree(self, src, dst, overwrite)

    def moveone(self, src, dst, overwrite):
        return moveone(self, src, dst, overwrite)

    def movetree(self, src, dst, overwrite):
        return movetree(self, src, dst, overwrite)

    def delone(self, uri):
        return delone(self, uri)

    def deltree(self, uri):
        return deltree(self, uri)

    def copy(self, src, dst):
        content = self._get_dav_getcontenttype(src)
        data = self.get_data(src)
        self.put(dst, data, content)
        return 201

    def copycol(self, src, dst):
        return self.mkcol(dst)

    def _get_dav_current_user_privilege_set(self, uri):
        dbname, dburi = self._get_dburi(uri)
        privileges = []
        if not dbname or not dburi:
            privileges = ['create', 'read', 'write', 'delete']
        else:
            pool = Pool(Transaction().database.name)
            try:
                Collection = pool.get('webdav.collection')
                privileges = Collection.current_user_privilege_set(
                    dburi, cache=LOCAL.cache)
            except KeyError:
                pass
            except Exception as exception:
                self._log_exception(exception)
                pass
        doc = domimpl.createDocument(None, 'privilege', None)
        privilege = doc.documentElement
        privilege.tagName = 'D:privilege'
        if 'create' in privileges:
            bind = doc.createElement('D:bind')
            privilege.appendChild(bind)
        if 'read' in privileges:
            read = doc.createElement('D:read')
            privilege.appendChild(read)
            read_acl = doc.createElement('D:read-acl')
            privilege.appendChild(read_acl)
        if 'write' in privileges:
            write = doc.createElement('D:write')
            privilege.appendChild(write)
            write_content = doc.createElement('D:write-content')
            privilege.appendChild(write_content)
            write_properties = doc.createElement('D:write-properties')
            privilege.appendChild(write_properties)
        if 'delete' in privileges:
            unbind = doc.createElement('D:unbind')
            privilege.appendChild(unbind)
        return privilege


TrytonDAVInterface.PROPS['DAV:'] = tuple(
    list(TrytonDAVInterface.PROPS['DAV:']) +
    ['current-user-privilege-set'])


class WebDAVAuthRequestHandler(WebDAVServer.DAVRequestHandler):

    def finish(self):
        WebDAVServer.DAVRequestHandler.finish(self)

        if not Transaction().connection:
            return
        Transaction().__exit__(None, None, None)
        # TODO: Check if vanished method below has to be replaced
        # if dbname:
        #     with Transaction().start(dbname, 0):
        #         Cache.resets(dbname)

    def parse_request(self):
        if not http.server.BaseHTTPRequestHandler.parse_request(self):
            return False

        authorization = self.headers.get('Authorization', '')
        if authorization:
            scheme, credentials = authorization.split()
            if scheme != 'Basic':
                self.send_error(501)
                return False
            credentials = decodebytes(credentials.encode()).decode()

            user, password = credentials.split(':', 2)

            if not self.get_userinfo(user, password, self.command):
                self.send_autherror(401, "Authorization Required")
                return False
        else:
            if not self.get_userinfo(None, None, self.command):
                self.send_autherror(401, "Authorization Required")
                return False
        return True

    def get_userinfo(self, user, password, command=''):
        path = urllib.parse.urlparse(self.path).path
        dbname = urllib.parse.unquote_plus(path.split('/', 2)[1])
        with Transaction().start(dbname, 0, close=True) as transaction:
            databases = transaction.database.list()
        if not dbname or dbname not in databases:
            return True
        if user:

            user = str(user)
            parameters = {'password': password}
            user = login(dbname, user, parameters, cache=False)

            if not user:
                return None
            user = int(user)
        else:
            url = urllib.parse.urlparse(self.path)
            query = urllib.parse.parse_qs(url.query)
            path = url.path[len(dbname) + 2:]
            if 'key' in query:
                key, = query['key']
                with Transaction().start(dbname, 0) as transaction:
                    database_list = Pool.database_list()
                    pool = Pool(dbname)
                    if dbname not in database_list:
                        pool.init()
                    Share = pool.get('webdav.share')
                    user = Share.get_login(key, command, path)
                    transaction.commit()
            if not user:
                return None

        Transaction().start(dbname, user, context={
            '_check_access': True,
        }, autocommit=True)
        # TODO: Check if vanished method below has to be replaced
        # Cache.clean(dbname)
        return user


class SecureWebDAVAuthRequestHandler(WebDAVAuthRequestHandler):

    def setup(self):
        self.request = SSLSocket(self.request)
        WebDAVAuthRequestHandler.setup(self)



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2013 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                 HEALTH PEDIATRICS_GROWTH_CHARTS package               #
#              health_pediatrics_growth_charts.py: main module          #
#########################################################################

from dateutil.relativedelta import relativedelta
from trytond.model import fields
from trytond.pool import PoolMeta

__all__ = ['PatientEvaluation']
__metaclass__ = PoolMeta


class PatientEvaluation(metaclass=PoolMeta):
    __name__ = 'gnuhealth.patient.evaluation'

    age_months = fields.Function(
        fields.Integer('Patient Age in Months'),
        'get_patient_age_months')

    def get_patient_age_months(self, name):
        if self.patient:
            if self.patient.dob:
                delta = relativedelta(self.evaluation_start, self.patient.dob)
                return delta.years * 12 + delta.months
        return None



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2014 Sebastian Marro <smarro@thymbra.com>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH REPORTING package                      #
#                 __init__.py: Package declaration file                 #
#########################################################################

from trytond.pool import Pool
from . import wizard
from . import report


def register():
    Pool.register(
        wizard.wizard_top_diseases.TopDiseases,
        wizard.wizard_top_diseases.OpenTopDiseasesStart,
        wizard.wizard_evaluations.OpenEvaluationsStart,
        wizard.wizard_summary_report.SummaryReportStart,
        wizard.wizard_evaluations.EvaluationsDoctor,
        wizard.wizard_evaluations.EvaluationsSpecialty,
        wizard.wizard_evaluations.EvaluationsSector,
        wizard.wizard_epidemics_report.EpidemicsReportStart,
        module='health_reporting', type_='model')
    Pool.register(
        wizard.wizard_top_diseases.OpenTopDiseases,
        wizard.wizard_evaluations.OpenEvaluations,
        wizard.wizard_summary_report.SummaryReport,
        wizard.wizard_epidemics_report.EpidemicsReport,
        module='health_reporting', type_='wizard')

    Pool.register(
        report.summary_report.InstitutionSummaryReport,
        report.epidemics_report.InstitutionEpidemicsReport,
        module='health_reporting', type_='report')



# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
# SPDX-FileCopyrightText: 2023 ITARA <info@itaraong.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                  HEALTH  SURGERY PROTOCOLS PACKAGE                    #
#              health_surgery_protocols.py: main module                 #
#########################################################################



# SPDX-FileCopyrightText: 2020-2021 National University of Entre Rios (UNER)
#                         School of Engineering
#                         <saludpublica@ingenieria.uner.edu.ar>
# SPDX-FileCopyrightText: 2020 Mario Puntin <mario@silix.com.ar>
# SPDX-FileCopyrightText: 2020-2024 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2020-2024 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH DENTISTRY package                      #
#                __init__.py: Package declaration file                  #
#########################################################################


from trytond.pool import Pool
from . import health_dentistry
from . import wizard
from . import report


def register():
    Pool.register(
        health_dentistry.PatientData,
        health_dentistry.DentistryTreatment,
        health_dentistry.DentistryProcedure,
        health_dentistry.TreatmentProcedure,
        wizard.patient_set_odontogram.SetOdontogramStart,
        wizard.load_procedure.LoadProcedureStart,
        module='health_dentistry', type_='model')
    Pool.register(
        wizard.patient_set_odontogram.SetOdontogram,
        wizard.load_procedure.LoadProcedure,
        module='health_dentistry', type_='wizard')
    Pool.register(
        report.procedures_report.DentistryProcedureReport,
        report.odontogram_report.Odontogram,
        module='health_dentistry', type_='report')



# SPDX-FileCopyrightText: 2008-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2011-2025 GNU Solidario <health@gnusolidario.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                      HEALTH STOCK SURGERY package                     #
#                 __init__.py: Package declaration file                 #
#########################################################################

from trytond.pool import Pool
from . import health_stock_surgery
from . import wizard


def register():
    Pool.register(
        health_stock_surgery.Move,
        health_stock_surgery.Surgery,
        wizard.wizard_create_surgery_stock_move.
        CreateSurgeryStockMoveInit,
        module='health_stock_surgery', type_='model')
    Pool.register(
        wizard.wizard_create_surgery_stock_move.
        CreateSurgeryStockMove,
        module='health_stock_surgery', type_='wizard')



# SPDX-FileCopyrightText: 2019-2022 Chris Zimmerman <chris@teffalump.com>
# SPDX-FileCopyrightText: 2021-2025 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2023 Patryk Rosik <p.rosik@stud.uni-hannover.de>
# SPDX-FileCopyrightText: 2023-2025 Feng Shu <tumashu@163.com>
# SPDX-FileCopyrightText: 2024-2025 Wei Zhao <wei.zhao@uclouvain.be>
# SPDX-FileCopyrightText: 2021-2025 GNU Solidario <health@gnusolidario.org>

#
# SPDX-License-Identifier: GPL-3.0-or-later
#########################################################################
#   Hospital Management Information System (HMIS) component of the      #
#                       GNU Health project                              #
#                   https://www.gnuhealth.org                           #
#########################################################################
#                         HEALTH ORTHANC package                        #
#                     health_orthanc.py: main module                    #
#########################################################################

"""
Core module of the Orthanc DICOM Server integration.

This module provides models for synchronization between an Orthanc DICOM
Server and the GNU Health HMIS. It provides methods to check when the
last synchronization took  place and makes them available to users via
appropriate GUI elements. Additionally,  there are methods to provide
hyperlinks to the corresponding studies and patients for
given Orthanc DICOM servers.
"""

from trytond.model import ModelView, ModelSQL, fields, Unique
from trytond.pool import Pool, PoolMeta
from trytond.exceptions import UserError
from pyorthanc import Orthanc
from datetime import datetime
from urllib.parse import urljoin
from lxml import etree
from trytond.i18n import gettext

import logging

try:
    from trytond.modules.health_imaging_worklist.health_imaging_worklist \
        import gnuhealth_org_root
except ImportError:
    gnuhealth_org_root = None


__all__ = [
    "View",
    "TestResult",
    "Patient",
    "PatientOrthancStudy",
    'StudySeries',
    'SeriesInstances',
    # DEPRECATED, Used to migrate date.
    "OrthancPatientDEPRECATED",
    "OrthancStudyDEPRECATED"
]

logger = logging.getLogger(__name__)


#
#  Adding widget "binary_dicom" to the server
#

class View(metaclass=PoolMeta):
    __name__ = 'ir.ui.view'

    @classmethod
    def get_rng(cls, type_):
        rng = super(View, cls).get_rng(type_)
        if type_ in ('form', 'list-form'):
            widgets = rng.xpath(
                '//ns:define/ns:optional/ns:attribute'
                '/ns:name[.="widget"]/following-sibling::ns:choice',
                namespaces={'ns': 'http://relaxng.org/ns/structure/1.0'})[0]
            subelem = etree.SubElement(
                widgets, '{http://relaxng.org/ns/structure/1.0}value')
            subelem.text = 'binary_dicom'
        return rng


class TestResult(metaclass=PoolMeta):
    __name__ = "gnuhealth.imaging.test.result"

    """
    Adds Orthanc imaging studies to imaging test result.
    """

    orthanc_studies = fields.One2Many(
        "gnuhealth.imaging_orthanc.study",
        "imaging_test", "Orthanc studies",
        readonly=True)

    # Deprecated in GH 5.0 . Use orthanc_studies.
    studies = fields.One2Many(
        "gnuhealth.orthanc.study",
        "imaging_test", "Orthanc studies",
        readonly=True)

    @classmethod
    def create(cls, vlist):
        Request = Pool().get('gnuhealth.imaging.test.request')
        vlist = [x.copy() for x in vlist]

        for values in vlist:
            request = Request.search(
                [("id", "=", values['request'])], limit=1)[0]

            studies = cls.find_orthanc_studies(request)

            if studies:
                values['orthanc_studies'] = [('add', [x.id for x in studies])]

        return super(TestResult, cls).create(vlist)

    @classmethod
    def find_orthanc_studies(cls, request):
        if (request and getattr(request, 'merge_id', None)
                and len(request.merge_id) > 0):
            Study = Pool().get('gnuhealth.imaging_orthanc.study')
            studies = Study.search(
                [("merge_id", "=", request.merge_id)])
            return studies


class Patient(metaclass=PoolMeta):
    __name__ = "gnuhealth.patient"

    """
    Adds Orthanc patients to the main patient data.
    """

    orthanc_patients = fields.One2Many(
        "gnuhealth.orthanc.patient",
        "patient", "Orthanc patients")

    orthanc_studies = fields.One2Many(
        'gnuhealth.imaging_orthanc.study',
        'patient', 'Orthanc Study')


class PatientOrthancStudy(ModelSQL, ModelView):
    'Patient Orthanc Study'
    __name__ = "gnuhealth.imaging_orthanc.study"

    patient = fields.Many2One(
        'gnuhealth.patient', 'Patient',
        help='Patient Name',
        readonly=False)

    imaging_test = fields.Many2One(
        "gnuhealth.imaging.test.result", "Result")

    date = fields.Char('Date', required=False, readonly=True)

    patient_name = fields.Char(
        'Orthanc Patient',
        required=True, readonly=True)

    patient_id = fields.Char("Patient ID", readonly=True)

    study_instance_UID = fields.Char(
        'Study UID',
        required=True, readonly=True)

    orthanc_UID = fields.Char(
        'Orthanc UID',
        readonly=True, required=True)

    merge_id = fields.Char(
        "Merge ID", readonly=True,
        help="Test result merge id, with it help, "
        "gnuhealth test result and orthanc study can be merged.")

    merge_comment = fields.Char(
        'Merge Comment')

    institution = fields.Char('Institution', readonly=True)
    performing_physician_name = fields.Char('Physician', readonly=True)

    series = fields.One2Many(
        'gnuhealth.imaging_orthanc.study_series',
        'study', 'Study Series')

    server = fields.Char('Orthanc Server', readonly=True, required=True)

    link_base_url = fields.Function(
        fields.Char(
            "Link Base URL",
            help="Base URL for links"),
        "get_link_base_url")

    ohif_viewer_link = fields.Function(
        fields.Char(
            "OHIF Viewer",
            help="Link to Orthanc OHIF Viewer"),
        "get_ohif_viewer_link")

    stone_viewer_link = fields.Function(
        fields.Char(
            "Stone Viewer",
            help="Link to Orthanc Stone Viewer"),
        "get_stone_viewer_link")

    gnuhealth_patient_name = fields.Function(
        fields.Char('Health Patient'),
        "get_gnuhealth_patient")

    notes = fields.Text(
        "Study notes",
        help='Extra Information',
        readonly=False)

    @classmethod
    def __setup__(cls):
        # Setup the patient orthanc study class with additional buttons
        # for deleting a study and selecting a viewer.
        super(PatientOrthancStudy, cls).__setup__()
        t = cls.__table__()

        cls._buttons.update({
            'delete_study': {}
        })
        cls._sql_constraints = [
            ('studyUID_unique', Unique(t, t.study_instance_UID, t.server),
             ("There is already a study with the same UID. "
              "Use the \"Get New studies\" action to get "
              "the latest list of studies from the Orthanc servers "
              "to see whether there is already a study with the same UID."))
        ]

        cls._order.insert(0, ('patient_name', 'ASC'))

    def get_gnuhealth_patient(self, name):
        """
        Retrieves the GNU patient with the given name.
        Returns:
            str or None: The name of the patient if found, None otherwise.
        """
        if self.patient is None:
            return None
        else:
            return self.patient.rec_name

    def get_link_base_url(self, name):
        pool = Pool()
        Config = pool.get('gnuhealth.orthanc.config')
        server_configs = Config.search([('domain', '=', self.server)])
        if len(server_configs) == 0:
            return self.server
        else:
            return server_configs[0].link_base_url

    def get_ohif_viewer_link(self, name):
        """
        Get the link for the OHIF viewer and study, based on the server,
        study_instance_UID, and orthanc_UID.
        """
        url = urljoin(self.link_base_url, (
            'ohif/viewer?' +
            f'StudyInstanceUIDs={self.study_instance_UID}'))
        return url

    def get_stone_viewer_link(self, name):
        """
        Get the link for the stone viewer and study, based on the
        server and study instance UID.
        """
        url = urljoin(self.link_base_url, (
            'stone-webviewer/index.html?' +
            f'study={self.study_instance_UID}'))
        return url

    @classmethod
    @ModelView.button
    def delete_study(cls, records):
        """
        Method to delete a study record from Orthanc server.
        :param records: List of study records to be deleted
        :return: 'reload' if the operation is successful
        :raises: UserError if there is an issue with the Orthanc server
        """
        records_to_delete = []
        try:
            Config = Pool().get('gnuhealth.orthanc.config')
            servers = Config.search([])
            for record in records:
                for conf_server in servers:
                    if conf_server.domain == record.server:
                        client = Orthanc(
                            url=conf_server.domain,
                            username=conf_server.user,
                            password=conf_server.password,
                            return_raw_response=True)
                        response = client.delete_studies_id(record.orthanc_UID)
                        if (200 <= response.status_code < 300 or
                                response.status_code == 404):
                            records_to_delete.append(record)
                        else:
                            raise UserError(
                                'Orthanc server returned HTTP code '
                                f'{response.status_code}, '
                                f'with content {response.text}',
                                description=(
                                    "Unable to delete Orthanc study. "
                                    "It may no longer exist or "
                                    "the Orthanc server "
                                    "could be in read-only mode."
                                    "Please review your "
                                    "Orthanc server configuration."))
        except Exception as exception:
            logger.error(
                'Delete Orthanc study exception: %s',
                exception,
                exc_info=True)
            raise UserError(str(exception))
        finally:
            cls.delete(records_to_delete)
            return 'reload'

    @classmethod
    def get_new_studies(cls):
        # Update the studies from the Orthanc server,
        # processing changes to studies, series, and instances.
        try:
            pool = Pool()
            Config = pool.get('gnuhealth.orthanc.config')
            server_configs = Config.search([])
            for server_config in server_configs:
                client = Orthanc(
                    url=server_config.domain,
                    username=server_config.user,
                    password=server_config.password)

                # get changes
                last_changed_index = (
                    server_config.last_changed_index
                    if server_config.last_changed_index is not None else -1)

                new_changes = client.get_changes(
                    {"since": last_changed_index, "limit": 100000})

                # process changes to studies
                new_orthanc_studyIDs = [
                    s['ID'] for s in new_changes['Changes']
                    if s['ChangeType'] == 'NewStudy'
                    or s['ChangeType'] == 'StableStudy']

                cls.create_or_update_studies_from_orthanc(
                    client, server_config.domain,
                    new_orthanc_studyIDs)

                # process changes to series
                new_orthanc_series_ids = [
                    s['ID'] for s in new_changes['Changes']
                    if s['ChangeType'] == 'NewSeries'
                    or s['ChangeType'] == 'StableSeries']

                cls.create_or_update_series_from_orthanc(
                    client, server_config.domain,
                    new_orthanc_series_ids)

                # process changes to instances
                new_orthanc_instance_ids = [
                    s['ID'] for s in new_changes['Changes']
                    if s['ChangeType'] == 'NewInstance'
                    or s['ChangeType'] == 'StableInstance']

                cls.create_or_update_instances_from_orthanc(
                    client, server_config.domain,
                    new_orthanc_instance_ids)

                # remember last processed change
                server_config.last_changed_index = new_changes['Last']
                Config.save([server_config])
        except Exception as exception:
            raise UserError(
                str(exception),
                description="Failed to get new changes, "
                "please check the Orthanc server")

    @classmethod
    def create_or_update_studies_from_orthanc(
            cls, client, server, orthanc_study_ids):
        """
        Create or update imaging_orthanc studies from Orthanc server "
        "based on the provided study IDs.
        """
        pool = Pool()
        Study = pool.get('gnuhealth.imaging_orthanc.study')
        for orthanc_study_id in orthanc_study_ids:
            orthanc_study = client.get_studies_id(orthanc_study_id)
            dicom_tags = orthanc_study['MainDicomTags']
            patient_main_dicom_tags = orthanc_study['PatientMainDicomTags']
            study_values = {}

            # Create or update?
            gh_study = Study.search(
                [('study_instance_UID', '=',
                  dicom_tags['StudyInstanceUID']),
                 ('server', '=', server)])

            if len(gh_study) == 0:
                study_values['patient'] = None

                study_values['date'] = (
                    dicom_tags['StudyDate']
                    if 'StudyDate' in dicom_tags else "")

                study_values['study_instance_UID'] = \
                    dicom_tags['StudyInstanceUID']

                study_values['orthanc_UID'] = orthanc_study['ID']

                study_values['patient_name'] = (
                    patient_main_dicom_tags['PatientName']
                    if 'PatientName' in patient_main_dicom_tags
                    else "")

                study_values['patient_id'] = (
                    patient_main_dicom_tags['PatientID']
                    if 'PatientID' in patient_main_dicom_tags
                    else "")

                study_values['institution'] = (
                    dicom_tags['InstitutionName']
                    if 'InstitutionName' in dicom_tags else "")

                study_values['performing_physician_name'] = (
                    dicom_tags['ReferringPhysicianName']
                    if 'ReferringPhysicianName' else "")

                study_values['server'] = server

                study_values['merge_id'] = cls.get_merge_id(
                    orthanc_study, server)

                result = cls.find_test_result(study_values)
                if result:
                    study_values["imaging_test"] = result.id
                    study_values["patient"] = \
                        result.patient and result.patient.id
                else:
                    study_values['patient'] = \
                        cls.find_gnuhealth_patient(
                            study_values['patient_id'])

                Study.create([study_values])
            else:
                # DICOM studies are immutable. Only the internal Orthanc ID can
                # change.
                study_values['orthanc_UID'] = orthanc_study['ID']
                Study.write(gh_study, study_values)

            cls.update_imaging_test_request(study_values)

    @classmethod
    def get_merge_id(cls, orthanc_study, server):
        prefix = gnuhealth_org_root
        dicom_tags = orthanc_study['MainDicomTags']

        if not prefix:
            return None

        # In most situations, we use 'StudyInstanceUID' to store merge
        # id.
        if (dicom_tags['StudyInstanceUID'] or '').startswith(prefix):
            return dicom_tags['StudyInstanceUID']

        # XXX: for imaging workstation's bugs, sometimes, we use other
        # study tags instead of 'StudyInstanceUID' to store merge id.
        for (k, v) in dicom_tags.items():
            if isinstance(v, str) and v.startswith(prefix):
                return v

    @classmethod
    def find_test_result(cls, entry):
        if entry and entry.get("merge_id") and len(entry.get("merge_id")) > 0:
            Result = Pool().get('gnuhealth.imaging.test.result')
            result = Result.search(
                [("merge_id", "=", entry.get("merge_id"))],
                limit=1)
            return (result and result[0])

    @classmethod
    def find_gnuhealth_patient(cls, patient_id):
        # The length of PUID is 9, see generate_puid method in
        # gnuhealth.py
        if patient_id and len(patient_id) >= 9:
            Patient = Pool().get('gnuhealth.patient')
            patient = Patient.search(
                [("puid", "=", patient_id)],
                limit=1)
            return (patient and patient[0])

    @classmethod
    def update_imaging_test_request(cls, entry):
        if entry and entry.get("merge_id") and len(entry.get("merge_id")) > 0:
            Request = Pool().get('gnuhealth.imaging.test.request')
            request = Request.search(
                [("merge_id", "=", entry.get("merge_id"))],
                limit=1)
            # If we fetch studies from orthanc successfully, we will
            # set worklist_status field of request to done, with the
            # help of this field, we can control worklist process or
            # not in worklist server.
            if len(request) > 0 and getattr(
                    request[0], 'worklist_status', None):
                Request.write(request, {'worklist_status': 'done'})

    @classmethod
    def create_or_update_series_from_orthanc(
            cls, client, server, orthanc_seriesIDs):
        """
        Create or update series from Orthanc "
        "in the imaging_orthanc study and series models.
        """

        pool = Pool()
        Study = pool.get('gnuhealth.imaging_orthanc.study')
        Series = pool.get('gnuhealth.imaging_orthanc.study_series')
        for orthanc_seriesID in orthanc_seriesIDs:
            orthanc_series = client.get_series_id(orthanc_seriesID)
            dicom_tags = orthanc_series['MainDicomTags']
            series_values = {}

            # Create or update?
            gh_series = Series.search(
                [('series_UID', '=', dicom_tags['SeriesInstanceUID']),
                 ('server', '=', server)])

            if len(gh_series) == 0:
                series_values['modality'] = (
                    dicom_tags['Modality']
                    if 'Modality' in dicom_tags else "")

                series_values['series_UID'] = dicom_tags['SeriesInstanceUID']
                series_values['orthanc_UID'] = orthanc_series['ID']

                series_values['series_description'] = (
                    dicom_tags['SeriesDescription']
                    if 'SeriesDescription' in dicom_tags else "")

                series_values['series_number'] = (
                    dicom_tags['SeriesNumber']
                    if 'SeriesNumber' in dicom_tags else "")

                gh_study = Study.search(
                    [('orthanc_UID', '=', orthanc_series['ParentStudy']),
                     ('server', '=', server)])

                if len(gh_study) == 0:
                    raise UserError(
                        "The study with the given Orthanc ID " +
                        "does not exist in the gnuhealth database")

                Study.write(
                    gh_study, {
                        'series': [
                            ('create', [series_values])]})
            else:
                # DICOM series are immutable. Only the internal Orthanc ID can
                # change.
                series_values['orthanc_UID'] = orthanc_series['ID']
                Series.write(gh_series, series_values)

    @classmethod
    def create_or_update_instances_from_orthanc(
            cls, client, server, orthanc_instance_ids):
        """
        Create or update instances from Orthanc in the GNU Health system.
        """
        pool = Pool()
        Series = pool.get('gnuhealth.imaging_orthanc.study_series')
        IInst = pool.get('gnuhealth.imaging_orthanc.series_instances')
        for orthanc_instance_id in orthanc_instance_ids:
            orthanc_instance = client.get_instances_id(orthanc_instance_id)
            dicom_tags = orthanc_instance['MainDicomTags']
            instance_values = {}

            # Create or update?
            gh_instance = IInst.search(
                [('sop_instance_UID', '=', dicom_tags['SOPInstanceUID']),
                 ('server', '=', server)])

            if len(gh_instance) == 0:
                instance_values['sop_instance_UID'] = \
                    dicom_tags['SOPInstanceUID']

                instance_values['orthanc_UID'] = orthanc_instance['ID']

                if ('InstanceNumber' in dicom_tags
                    and dicom_tags['InstanceNumber'] != ""
                        and dicom_tags['InstanceNumber'] is not None):
                    instance_values['instance_number'] = int(
                        float(dicom_tags['InstanceNumber']))
                else:
                    instance_values['instance_number'] = 0

                instance_values['image_position_patient'] = (
                    dicom_tags['ImagePositionPatient']
                    if 'ImagePositionPatient' in dicom_tags
                    else "").replace("\\", ", ")

                gh_series = Series.search(
                    [('orthanc_UID', '=', orthanc_instance['ParentSeries']),
                     ('server', '=', server)])

                if len(gh_series) == 0:
                    raise UserError(
                        "The series with the given Orthanc ID "
                        "does not exist in the gnuhealth database")

                Series.write(
                    gh_series, {
                        'instances': [
                            ('create', [instance_values])]})
            else:
                # DICOM instances are immutable. Only the internal Orthanc ID
                # can change.
                instance_values['orthanc_UID'] = orthanc_instance['ID']
                IInst.write(gh_instance, instance_values)

    @classmethod
    def full_synchronize(cls):
        """
        Updates the studies in the gnuhealth database by "
        "fetching the studies from the Orthanc servers.
        This class method fetches all the studies that are "
        "already in the gnuhealth database and compares them
        with the studies from the Orthanc servers. It then creates new studies
        in the gnuhealth database if there are any studies in the Orthanc
        servers that are not already present in the gnuhealth database.
        """
        try:
            pool = Pool()
            Study = pool.get('gnuhealth.imaging_orthanc.study')
            Series = pool.get('gnuhealth.imaging_orthanc.study_series')
            # IInst = pool.get('gnuhealth.imaging_orthanc.series_instances')

            # Get all studies that are already in gnuhealth
            gh_studies = Study.search([])
            # Get studies from Orthanc servers
            Config = pool.get('gnuhealth.orthanc.config')
            servers = Config.search([])
            for server in servers:
                client = Orthanc(
                    url=server.domain,
                    username=server.user,
                    password=server.password)
                orthanc_studies = client.get_studies({'expand': True})
                for orthanc_study in orthanc_studies:
                    # Create study in gnuhealth if it does not exist
                    dicom_tags = orthanc_study['MainDicomTags']
                    iuid = dicom_tags['StudyInstanceUID']
                    gh_study = [
                        s for s in gh_studies
                        if s.study_instance_UID == iuid
                        and s.server == server.domain]

                    if len(gh_study) == 0:
                        dicom_tags = orthanc_study['MainDicomTags']

                        patient_main_dicomTags = \
                            orthanc_study['PatientMainDicomTags']

                        study_values = {}
                        study_values['patient'] = None

                        study_values['date'] = (
                            dicom_tags['StudyDate']
                            if 'StudyDate' in dicom_tags else "")

                        study_values['study_instance_UID'] = \
                            dicom_tags['StudyInstanceUID']

                        study_values['orthanc_UID'] = orthanc_study['ID']

                        study_values['patient_name'] = (
                            patient_main_dicomTags['PatientName']
                            if 'PatientName' in patient_main_dicomTags else "")

                        study_values['patient_id'] = (
                            patient_main_dicomTags['PatientID']
                            if 'PatientID' in patient_main_dicomTags else "")

                        study_values['institution'] = (
                            dicom_tags['InstitutionName']
                            if 'InstitutionName' in dicom_tags else "")

                        study_values['performing_physician_name'] = (
                            dicom_tags['ReferringPhysicianName']
                            if 'ReferringPhysicianName' else "")

                        study_values['server'] = server.domain

                        study_values['merge_id'] = cls.get_merge_id(
                            orthanc_study, server)

                        result = cls.find_test_result(study_values)
                        if result:
                            study_values["imaging_test"] = result.id
                            study_values["patient"] = \
                                result.patient and result.patient.id
                        else:
                            study_values['patient'] = \
                                cls.find_gnuhealth_patient(
                                    study_values['patient_id'])

                        logger.error("Creating study")
                        gh_study = Study.create([study_values])

                        cls.update_imaging_test_request(study_values)

                    gh_study = gh_study[0]

                    orthanc_seriesIDs = orthanc_study['Series']
                    for orthanc_seriesID in orthanc_seriesIDs:
                        orthanc_series = client.get_series_id(orthanc_seriesID)
                        # Create series in gnuhealth if it does not exist
                        gh_series = [
                            s for s in gh_study.series
                            if s.orthanc_UID == orthanc_seriesID]
                        if len(gh_series) == 0:
                            series_values = {}
                            series_main_dicomTags = \
                                orthanc_series['MainDicomTags']

                            series_values['modality'] = (
                                series_main_dicomTags['Modality']
                                if 'Modality' in series_main_dicomTags else "")

                            series_values['series_UID'] = \
                                series_main_dicomTags['SeriesInstanceUID']

                            series_values['orthanc_UID'] = orthanc_series['ID']

                            series_values['series_description'] = (
                                series_main_dicomTags['SeriesDescription']
                                if 'SeriesDescription'
                                in series_main_dicomTags else "")

                            series_values['series_number'] = (
                                series_main_dicomTags['SeriesNumber']
                                if 'SeriesNumber' in series_main_dicomTags
                                else "")

                            series_values['study'] = gh_study
                            logger.error("Creating series")
                            gh_series = Series.create([series_values])
                            Study.write(
                                [gh_study], {
                                    'series': [
                                        ('add', [
                                            gh_series[0].id])]})
                        gh_series = gh_series[0]

                        instance_values_to_create = []
                        for orthanc_instanceID in orthanc_series['Instances']:
                            # Create instance in gnuhealth if it does not exist
                            gh_instance = [
                                s for s in gh_series.instances
                                if s.orthanc_UID == orthanc_instanceID]

                            if len(gh_instance) == 0:
                                orthanc_instance = client.get_instances_id(
                                    orthanc_instanceID)
                                instance_values = {}
                                instance_main_dicomTags = \
                                    orthanc_instance['MainDicomTags']

                                instance_values['sop_instance_UID'] = \
                                    instance_main_dicomTags['SOPInstanceUID']

                                instance_values['orthanc_UID'] = \
                                    orthanc_instance['ID']

                                itags = instance_main_dicomTags
                                if ('InstanceNumber' in itags and
                                    itags['InstanceNumber'] != '' and
                                        itags['InstanceNumber'] is not None):
                                    instance_values['instance_number'] = int(
                                        float(itags['InstanceNumber']))
                                else:
                                    instance_values['instance_number'] = 0
                                instance_values['image_position_patient'] = (
                                    itags['ImagePositionPatient']
                                    if 'ImagePositionPatient' in itags
                                    else "").replace("\\", ", ")
                                instance_values_to_create.append(
                                    instance_values)
                        logger.error(
                            "Creating " +
                            str(len(instance_values_to_create)) +
                            " instances")
                        Series.write(
                            [gh_series], {
                                'instances': [
                                    ('create', instance_values_to_create)]})
        except Exception as exception:
            raise UserError(
                str(exception),
                description="Failed to update studies, "
                "please check the Orthanc server")


class StudySeries(ModelSQL, ModelView):
    'Study Series'
    __name__ = 'gnuhealth.imaging_orthanc.study_series'

    study = fields.Many2One(
        'gnuhealth.imaging_orthanc.study', 'Study',
        help='Patient study series',
        readonly=True,
        required=True,
        ondelete='CASCADE')

    orthanc_UID = fields.Char('Orthanc UID', readonly=True, required=True)
    series_number = fields.Char('Series No.', readonly=True)

    series_description = fields.Char(
        'Description',
        readonly=True, required=False)

    series_UID = fields.Char('Series UID', readonly=True, required=True)
    modality = fields.Char('Modality', required=True, readonly=True)

    server = fields.Function(
        fields.Char(
            "Server",
            readonly=True, required=True),
        "get_study_server",
        searcher='search_study_server')

    stone_viewer_link = fields.Function(
        fields.Char(
            "Stone Viewer",
            help="Link to Orthanc Stone Viewer"),
        "get_stone_viewer_link")

    notes = fields.Text(
        "Series notes",
        help='Extra Information',
        readonly=False)

    instances = fields.One2Many(
        'gnuhealth.imaging_orthanc.series_instances',
        'series',
        'Series Instance')

    @classmethod
    def __setup__(cls):
        """
        A description of the entire function, "
        "its parameters, and its return types.
        """
        super(StudySeries, cls).__setup__()
        cls._buttons.update({
            'delete_series': {},
        })

    def get_stone_viewer_link(self, name):
        """
        Get the link for the stone viewer and study, based on the
        server and study instance UID.
        """
        # https://orthanc.uclouvain.be/demo/stone-webviewer/index.html?study=1.2.840.113745.101000.1008000.38179.6792.6324567&series=1.3.12.2.1107.5.1.4.36085.2.0.517109821292363

        url = urljoin(self.study.link_base_url, (
            'stone-webviewer/index.html?' +
            f'study={self.study.study_instance_UID}' +
            f'&series={self.series_UID}'))

        return url

    def get_study_server(self, name):
        """
        Get the study server.

        Parameters:
            name (str): The name of the study.

        Returns:
            study_server: The server for the given study.
        """
        return self.study.server

    @classmethod
    def search_study_server(cls, name, clause):
        """
        Perform a search on the study server "
        "based on the provided name and clause.
        """
        res = []
        value = clause[2]
        res.append(('study.server', clause[1], value))
        return res

    def get_study_patient(self, name):
        """
        Returns the name of the patient associated with the given study.

        Parameters:
            name (str): The name of the study.

        Returns:
            str: The name of the patient associated with the study.
        """
        return self.study.patient.name

    @classmethod
    def delete(cls, seriess):
        """
        Delete a list of series and their associated studies "
        "if they have no more series.
        """
        studies = [s.study for s in seriess]
        # call original delete
        super(StudySeries, cls).delete(seriess)
        # if the study has no more series, delete the study, too
        Study = Pool().get('gnuhealth.imaging_orthanc.study')
        for study in studies:
            if study and ((study.series is None) or len(study.series) == 0):
                Study.delete([study])

    @classmethod
    @ModelView.button
    def delete_series(cls, records):
        """
        A class method to delete series from the Orthanc server.
        Takes a list of records as input.
        Returns 'reload' on successful deletion.
        Raises UserError on failure with an appropriate error message.
        """
        records_to_delete = []
        try:
            Config = Pool().get('gnuhealth.orthanc.config')
            servers = Config.search([])
            for record in records:
                for conf_server in servers:
                    if (conf_server.domain == record.study.server):
                        client = Orthanc(
                            url=conf_server.domain,
                            username=conf_server.user,
                            password=conf_server.password,
                            return_raw_response=True)

                        response = client.delete_series_id(record.orthanc_UID)
                        if (200 <= response.status_code < 300 or
                                response.status_code == 404):
                            records_to_delete.append(record)
                        else:
                            raise UserError(
                                'Orthanc server returned HTTP code'
                                f'{response.status_code}, '
                                f'with content {response.text}',
                                description=(
                                    "Unable to delete Orthanc study"
                                    "series. It may no longer exist or "
                                    "the Orthanc server could be in "
                                    "read-only mode. "
                                    "Please review your "
                                    "Orthanc server configuration."))
        except Exception as exception:
            logger.error(
                'Delete study series exception: %s',
                exception,
                exc_info=True)
            raise UserError(str(exception))
        finally:
            cls.delete(records_to_delete)
        return 'reload'

#
# All instances in the series of patient's image study.
#


class SeriesInstances(ModelSQL, ModelView):
    'Series Instance'
    __name__ = 'gnuhealth.imaging_orthanc.series_instances'

    series = fields.Many2One(
        'gnuhealth.imaging_orthanc.study_series',
        'Series',
        help='Study series instance',
        readonly=True,
        required=True,
        ondelete='CASCADE')

    sop_instance_UID = fields.Char(
        'SOP Instance UID',
        readonly=True,
        required=True)

    orthanc_UID = fields.Char('Orthanc UID', readonly=True, required=True)

    instance_number = fields.Integer(
        'Instance Number',
        readonly=True,
        required=False)

    image_position_patient = fields.Char(
        'Image Position', readonly=True, required=False)

    server = fields.Function(
        fields.Char(
            "Server",
            readonly=True,
            required=True),
        "get_study_server",
        searcher="search_study_server")

    image = fields.Function(
        fields.Binary("Image"),
        'get_image',
        loading='lazy')

    @classmethod
    def __setup__(cls):
        """
        Set up the SeriesInstances class.
        """
        super(SeriesInstances, cls).__setup__()

        cls._buttons.update({
            'attach_image_to_imaging_test_report': {}})

        cls._order.insert(0, ('instance_number', 'ASC'))

    def get_study_server(self, name):
        """
        Get the study server for a given name.

        :param name: The name of the study.
        :return: The server associated with the study.
        """
        return self.series.server

    @classmethod
    def search_study_server(cls, name, clause):
        """
        A class method that searches the study server "
        "based on a given name and clause.
        """
        res = []
        value = clause[2]
        res.append(('series.server', clause[1], value))
        return res

    def get_image(self, name):
        """
        A method to retrieve an image from an Orthanc server.

        :return: The image data in PNG format if successful, None otherwise.
        """
        try:
            Config = Pool().get('gnuhealth.orthanc.config')
            servers = Config.search([])
            for conf_server in servers:
                if conf_server.domain == self.server:
                    logger.error(conf_server.domain)
                    client = Orthanc(
                        url=conf_server.domain,
                        username=conf_server.user,
                        password=conf_server.password,
                        return_raw_response=True)
                    response = client.get_instances_id_frames_frame_rendered(
                        0, self.orthanc_UID, headers={'Accept': 'image/png'})
                    if 200 <= response.status_code < 300:
                        image_data = response.read()
                        return image_data
                    else:
                        raise UserError(
                            "Orthanc server returned HTTP code"
                            f"{response.status_code},"
                            f" with content {response.text}")
            return None
        except Exception as exception:
            logger.error('Get image of instance: %s', exception, exc_info=True)
            return None

    @classmethod
    @ModelView.button
    def attach_image_to_imaging_test_report(cls, records):
        Config = Pool().get('gnuhealth.orthanc.config')
        Attachment = Pool().get('ir.attachment')

        servers = Config.search([])

        for record in records:
            imaging_test = record.series.study.imaging_test
            image_data = None

            if imaging_test:
                for conf_server in servers:
                    if (conf_server.domain == record.server):
                        client = Orthanc(
                            url=conf_server.domain,
                            username=conf_server.user,
                            password=conf_server.password,
                            return_raw_response=True)
                        response = \
                            client.get_instances_id_frames_frame_rendered(
                                0, record.orthanc_UID,
                                headers={'Accept': 'image/png'})

                        if 200 <= response.status_code < 300:
                            image_data = response.read()

                            label_instance_number = gettext(
                                'health_orthanc.msg_label_instance_number')

                            label_image_position = gettext(
                                'health_orthanc.msg_label_image_position')

                            description = (
                                f"{label_instance_number} "
                                f"{record.instance_number}\n"
                                f"{label_image_position}\n"
                                f"... ({record.image_position_patient})")

                            attachment = Attachment.search(
                                [('resource', '=', imaging_test),
                                 ('name', '=', record.sop_instance_UID)])

                            if attachment:
                                Attachment.write(attachment, {
                                    'data': image_data,
                                    'description': description})
                            else:
                                Attachment.create([{
                                    'name': record.sop_instance_UID,
                                    'data': image_data,
                                    'description': description,
                                    'resource': imaging_test}])
                        else:
                            raise UserError(
                                "Orthanc server returned HTTP code"
                                f"{response.status_code},"
                                f" with content {response.text}")
            else:
                raise UserError(
                    gettext(
                        'health_orthanc.msg_find_no_gnuhealth_imaging_test'))

        return "reload"

# ----------------------------------------------------------------------------
# WARN: The following code will be deprecated in GH 5.0
# ----------------------------------------------------------------------------


class OrthancPatientDEPRECATED(ModelSQL, ModelView):
    """Orthanc patient information"""
    """
    Defines an Orthanc Patient.

    This class defines the ``OrthancPatient``. It provides methods to update
    existing patients or add new patients from the Orthanc DICOM server.
    Additionally, it allows to extract DICOM tags and automatically generates
    hyperlinks to the corresponding patients.

    :param ModelSQL: Inherit from the Tryton ModelSQL class for SQL
                     database operations.
    :type ModelSQL: class: ``trytond.model.ModelSQL``

    :param ModelView: Inherit from the Tryton ModelView class for user
                      interface operations.
    :type ModelView: class: ``trytond.model.ModelView``

    :var __name__: The unique name ``gnuhealth.orthanc.patient`` of the model.
    :vartype __name__: str

    :var patient: Local linked patient from Orthanc into GNU Health HMIS.
    :vartype patient: class: ``trytond.model.field.Many2One``

    :var name: Name of the patient. Read-only.
    :vartype name: class: ``trytond.model.field.Char``

    :var bd: Birth date of the patient. Read-only.
    :vartype bd: class: ``trytond.model.field.Date``

    :var ident: Unique ID for a patient based on the Patient ID DICOM Tag.
                Read-only.
    :vartype ident: class: ``trytond.model.field.Char``

    :var uuid: SHA-1 Hash of ``ident``. Read-only and Required.
    :vartype uuid: class: ``trytond.model.field.Char``

    :var studies: List of Orthanc studies directly related to the patient.
                  Read-only.
    :vartype studies: class: ``trytond.model.field.One2Many``

    :var server:  A field to specify the server. Required.
    :vartype server: class: ``trytond.model.field.Many2One``

    :var link: Link to the patient in the Orthanc Explorer.
    :vartype link: class: ``trytond.model.field.Char``
    """

    __name__ = "gnuhealth.orthanc.patient"

    patient = fields.Many2One(
        "gnuhealth.patient", "Patient", help="Local linked patient"
    )
    name = fields.Char("PatientName", readonly=True)
    bd = fields.Date("Birthdate", readonly=True)
    ident = fields.Char("PatientID", readonly=True)
    uuid = fields.Char("PatientUUID", readonly=True, required=True)

    studies = fields.One2Many(
        "gnuhealth.orthanc.study", "patient", "Studies", readonly=True
    )

    server = fields.Many2One(
        "gnuhealth.orthanc.config", "Server", readonly=True)
    link = fields.Function(
        fields.Char(
            "URL", help="Link to patient in Orthanc Explorer"), "get_link"
    )

    def get_link(self, name):
        """
        Return a link to the Orthanc patient with the specified uuid in
        the Orthanc explorer.

        :param name: Label of the patient to get the link for.
        :type name: str

        :return: URL to the Orthanc patient in the Orthanc Explorer.
        :rtype: str
        """

        pre = "".join([self.server.domain.rstrip("/"), "/"])
        add = "app/explorer.html#patient?uuid={}".format(self.uuid)
        return urljoin(pre, add)

    @classmethod
    def __setup__(cls):
        """
        Set up the ``OrthancPatient`` class for database access.

        This method is a class method that initializes various properties
        and constraints of the ``OrthancPatient`` model. It sets up a SQL
        constraint to ensure that the ``server`` and  ``uuid`` column
        are unique.
        """

        super().__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            (
                "uuid_unique",
                Unique(t, t.server, t.uuid),
                "UUID must be unique for a given server",
            )
        ]

    @staticmethod
    def get_info_from_dicom(patients):
        """
        Extract patient information from DICOM data for writing to database.

        :param patients: List of DICOM data for patients.
        :type patients: list

        :return: List of dictionaries with patient information.
        :rtype: list
        """

        data = []
        for patient in patients:
            try:
                bd = datetime.strptime(
                    patient["MainDicomTags"]["PatientBirthDate"], "%Y%m%d"
                ).date()

            except Exception:
                logger.exception(
                    "Invalid date format. Please provide the date in "
                    "the format %Y%m%d"
                )
                bd = None
            data.append(
                {
                    "name": patient.get("MainDicomTags").get("PatientName"),
                    "bd": bd,
                    "ident": patient.get("MainDicomTags").get("PatientID"),
                    "uuid": patient.get("ID"),
                }
            )
        return data

    @classmethod
    def update_patients(cls, patients, server):
        """
        Update patients with new information from DICOM files.

        :param patients: A list of patient data in DICOM format.
        :type patients: list

        :param server: The server to update the patients on.
        :type server: str
        """

        entries = cls.get_info_from_dicom(patients)
        updates = []
        for entry in entries:
            try:
                patient = cls.search(
                    [("uuid", "=", entry["uuid"]),
                     ("server", "=", server)], limit=1
                )[0]
                patient.party = entry["name"]
                patient.bd = entry["bd"]
                patient.ident = entry["ident"]
                # don't update unless no patient attached
                if not patient.patient:
                    try:
                        g_patient = Patient.search(
                            [("puid", "=", entry["ident"])], limit=1
                        )[0]
                        patient.patient = g_patient
                        logger.info(
                            "New Matching PUID found for {}".
                            format(entry["ident"])
                        )
                    except IndexError:
                        logger.warning(
                            "No patient from GNU Health HMIS attached")
                updates.append(patient)
                logger.info("Updating patient {}".format(entry["uuid"]))
            except IndexError:
                continue
                logger.warning("Unable to update patient {}".
                               format(entry["uuid"]))
        cls.save(updates)

    @classmethod
    def create_patients(cls, patients, server):
        """
        Create patients with information from DICOM files.

        :param patients: A list of patient data in DICOM format.
        :type patients: list

        :param server: The server to update the patients on.
        :type server: str
        """

        pool = Pool()
        Patient = pool.get("gnuhealth.patient")

        entries = cls.get_info_from_dicom(patients)
        for entry in entries:
            try:
                g_patient = Patient.search(
                    [("puid", "=", entry["ident"])], limit=1)[0]
                logger.info("Matching PUID found for {}".format(entry["uuid"]))
            except IndexError:
                g_patient = None
            entry["server"] = server
            entry["patient"] = g_patient
        cls.create(entries)


class OrthancStudyDEPRECATED(ModelSQL, ModelView):
    """
    DEPRECATED IN GH 5.0 by PatientOrthancStudy
    ("gnuhealth.imaging_orthanc.study")

    Defines an Orthanc Study.

    This class defines the ``OrthancStudy``. It provides methods to update
    existing studies or add new studies to the Orthanc DICOM server.
    Additionally, it allows to extract DICOM tags and automatically generates
    hyperlinks to the corresponding studies.

    :param ModelSQL: Inherit from the Tryton ModelSQL class for SQL
                     database operations.
    :type ModelSQL: class: ``trytond.model.ModelSQL``

    :param ModelView: Inherit from the Tryton ModelView class for user
                      interface operations.
    :type ModelView: class: ``trytond.model.ModelView``

    :var __name__: The unique name `gnuhealth.orthanc.study` of the model.
    :vartype __name__: str

    :var patient: Local patient of Orthanc conncted to a study. Read-only.
    :vartype patient: class: ``trytond.model.fields.Many2One``

    :var uuid: SHA-1 Hash of the PatientID tag (0010,0020) and their
               StudyInstanceUID tag
        (0020,000d). Read-only and Required.
    :vartype uuid: class: ``trytond.model.fields.Char``

    :var description: Description of the study conducted. Read-only.
    :vartype description: class: ``trytond.model.fields.Char``

    :var date: Date on which the study was conducted. Read-only.
    :vartype date: class: ``trytond.model.fields.Date``

    :var ident: ID of a study based on the Study ID DICOM Tag. Read-only.
    :vartype ident: class: ``trytond.model.fields.Char``

    :var institution: Institution at which the study was conducted. Read-only.
    :vartype institution: class: ``trytond.model.fields.Char``

    :var ref_phys: The referring physician. Read-only.
    :vartype ref_phys: class: ``trytond.model.fields.Char``

    :var req_phys: The requesting physician. Read-only.
    :vartype req_phys: class: ``trytond.model.fields.Char``

    :var server: Server on which the study is located. Read-only.
    :vartype server: class: ``trytond.model.fields.Many2One``

    :var ohif_viewer_link: Link to study in OHIF Viewer.
    :vartype link: class: ``trytond.model.fields.Char``

    :var stone_viewer_link: Link to study in Stone Viewer.
    :vartype link: class: ``trytond.model.fields.Char``

    :var orthanc_explorer_link: Link to study in Orthanc Explorer.
    :vartype link: class: ``trytond.model.fields.Char``

    :var imaging_test: Corresponding request from GNU Health HMIS.
    :vartype imaging_test: class: ``trytond.model.fields.Many2One``
    """

    __name__ = "gnuhealth.orthanc.study"

    patient = fields.Many2One(
        "gnuhealth.orthanc.patient", "Patient", readonly=True)

    uuid = fields.Char("UUID", readonly=True, required=True)
    description = fields.Char("Description", readonly=True)
    date = fields.Date("Date", readonly=True)
    ident = fields.Char("ID", readonly=True)
    instance_uid = fields.Char("InstanceUID", readonly=True)
    requested_procedure_id = fields.Char(
        "RequestedProcedureID", readonly=True
    )
    merge_id = fields.Char(
        "Merge ID", readonly=True,
        help="Test result merge id, with it help, "
        "gnuhealth test result and orthanc study can be merged."
    )
    institution = fields.Char(
        "Institution", readonly=True,
        help="Imaging center where study was undertaken"
    )
    ref_phys = fields.Char("Referring Physician", readonly=True)
    req_phys = fields.Char("Requesting Physician", readonly=True)
    server = fields.Many2One(
        "gnuhealth.orthanc.config", "Server", readonly=True)

    ohif_viewer_link = fields.Function(
        fields.Char(
            "OHIF Viewer", help="Link to study in OHIF Viewer."),
        "get_ohif_viewer_link")

    stone_viewer_link = fields.Function(
        fields.Char(
            "Stone Viewer", help="Link to study in Stone Viewer."),
        "get_stone_viewer_link")

    orthanc_explorer_link = fields.Function(
        fields.Char(
            "Orthanc Explorer", help="Link to study in Orthanc Explorer."),
        "get_orthanc_explorer_link")

    def get_ohif_viewer_link(self, name):
        """
        Return a link to the Orthanc study with the specified uuid in the
        OHIF viewer.

        :param name: Label of the study to get the link for.
        :type name: str

        :return: URL to the Orthanc study in OHIF viewer.
        :rtype: str
        """

        pre = "".join([self.server.domain.rstrip("/"), "/"])
        add = "ohif/viewer?url=../studies/{}/ohif-dicom-json".format(
            self.uuid)
        return urljoin(pre, add)

    def get_stone_viewer_link(self, name):
        """
        Return a link to the Orthanc study with the specified uuid in the
        Stone Viewer.

        :param name: Label of the study to get the link for.
        :type name: str

        :return: URL to the Orthanc study in Stone Viewer.
        :rtype: str
        """

        pre = "".join([self.server.domain.rstrip("/"), "/"])
        add = "stone-webviewer/index.html?study={}".format(
            self.instance_uid)
        return urljoin(pre, add)

    def get_orthanc_explorer_link(self, name):
        """
        Return a link to the Orthanc study with the specified uuid in the
        Orthanc explorer.

        :param name: Label of the study to get the link for.
        :type name: str

        :return: URL to the Orthanc study in the Orthanc Explorer.
        :rtype: str
        """

        pre = "".join([self.server.domain.rstrip("/"), "/"])
        add = "app/explorer.html#study?uuid={}".format(self.uuid)
        return urljoin(pre, add)

    imaging_test = fields.Many2One("gnuhealth.imaging.test.result", "Study")

    @classmethod
    def __setup__(cls):
        """
        Set up the ``OrthancStudy`` class for database access.

        This method is a class method that initializes various
        properties and constraints of the ``OrthancStudy`` model.
        It sets up a SQL constraint to ensure that the ``server`` and
        ``uuid`` columns are unique.
        """

        super().__setup__()
        t = cls.__table__()
        cls._sql_constraints = [
            (
                "uuid_unique",
                Unique(t, t.server, t.uuid),
                "UUID must be unique for a given server",
            )
        ]

    def get_rec_name(self, name):
        """
        Return name of actual study.

        :param name: The name of the record.
        :type name: str

        :return: A string representing the display name of the record.
        :rtype: str

        .. note:: This method follows the Tryton Syntax.
            It is the getter function for the field ``rec_name``.
        .. seealso:: classmethod: ``ModelStorage.get_rec_name` from
            class: ``trytond.model.ModelStorage``
        """

        return ": ".join((self.ident or self.uuid, self.description or ""))

    @classmethod
    def get_info_from_dicom(cls, studies, server):
        """
        Extract study  information from DICOM data for writing to database.

        :param studies: List of DICOM data for studies.
        :type studies: list

        :return: List of dictionaries with study information.
        :rtype: list
        """

        data = []

        for study in studies:
            try:
                date = datetime.strptime(
                    study["MainDicomTags"]["StudyDate"], "%Y%m%d"
                ).date()
            except Exception:
                logger.exception(
                    "Invalid date format. Please provide the date in "
                    "the format %Y%m%d"
                )
                date = None
            try:
                description = \
                    study["MainDicomTags"]["RequestedProcedureDescription"]
            except KeyError:
                logger.warning("No description provided")
                description = None

            entry = {
                "parent_patient": study["ParentPatient"],
                "uuid": study["ID"],
                "description": description,
                "date": date,
                "ident": study.get("MainDicomTags").get("StudyID"),
                "instance_uid": study.get("MainDicomTags").get(
                    "StudyInstanceUID"),
                "requested_procedure_id": study.get("MainDicomTags").get(
                    "RequestedProcedureID"),
                "institution": study.get("MainDicomTags").get(
                    "InstitutionName"),
                "ref_phys": study.get("MainDicomTags").get(
                    "ReferringPhysicianName"
                ),
                "req_phys": study.get(
                    "MainDicomTags").get("RequestingPhysician")
            }

            entry['merge_id'] = cls.get_merge_id(entry, server)

            data.append(entry)

        return data

    @classmethod
    def get_merge_id(cls, entry, server):
        prefix = gnuhealth_org_root

        # In most situations, we use 'StudyInstanceUID' to store merge
        # id.
        if (entry['instance_uid'] or '').startswith(prefix):
            return entry['instance_uid']

        # XXX: for imaging workstation's bugs, sometimes, we use other
        # study tags instead of 'StudyInstanceUID' to store merge id.
        for (k, v) in entry.items():
            if isinstance(v, str) and v.startswith(prefix):
                return v

        # XXX: for imaging workstation's bugs, sometimes, we use
        # 'PatientID' tag to store merge id.
        Patient = Pool().get("gnuhealth.orthanc.patient")
        patient = Patient.search(
            [("uuid", "=", entry["parent_patient"]),
             ("server", "=", server)],
            limit=1)[0]
        if patient:
            if patient.ident:
                if patient.ident.startswith(prefix):
                    return patient.ident

    @classmethod
    def update_studies(cls, studies, server):
        """
        Update studies  with new information from DICOM files.

        :param studies: A list of studies data in DICOM format.
        :type studies: list

        :param server: The server to update the studies on.
        :type server: str
        """

        entries = cls.get_info_from_dicom(studies, server)
        updates = []
        for entry in entries:
            try:
                study = cls.search(
                    [("uuid", "=", entry["uuid"]),
                        ("server", "=", server)], limit=1
                )[0]

                result = cls.find_test_result(entry)

                study.description = entry["description"]
                study.date = entry["date"]
                study.ident = entry["ident"]
                study.instance_uid = entry["instance_uid"]
                study.merge_id = entry["merge_id"]
                study.institution = entry["institution"]
                study.ref_phys = entry["ref_phys"]
                study.req_phys = entry["req_phys"]
                if result:
                    study.imaging_test = result.id
                updates.append(study)
                logger.info("Updating study {}".format(entry["uuid"]))
            except IndexError:
                continue
                logger.warning(
                    "Unable to update study {}".format(entry["uuid"]))
        cls.save(updates)

    @classmethod
    def find_test_result(cls, entry):
        if entry and entry["merge_id"] and len(entry["merge_id"]) > 0:
            Result = Pool().get('gnuhealth.imaging.test.result')
            result = Result.search(
                [("merge_id", "=", entry["merge_id"])],
                limit=1)
            return (result and result[0])

    @classmethod
    def create_studies(cls, studies, server):
        """
        Create studies with information from DICOM files.

        :param studies: A list of studies data in DICOM format.
        :type studies: list

        :param server: The server to update the patients on.
        :type server: str
        """

        pool = Pool()
        Patient = pool.get("gnuhealth.orthanc.patient")

        entries = cls.get_info_from_dicom(studies, server)
        for entry in entries:
            try:
                patient = Patient.search(
                    [("uuid", "=",
                      entry["parent_patient"]), ("server", "=", server)],
                    limit=1,
                )[0]
            except IndexError:
                patient = None
                logger.warning(
                    "No parent patient found for study {}".format(entry["ID"])
                )
            entry.pop("parent_patient")  # remove non-model entry
            entry["server"] = server
            entry["patient"] = patient
            result = cls.find_test_result(entry)
            if result:
                entry["imaging_test"] = result.id

        cls.create(entries)
